"""
Classes for handling raster data.
"""

from . import grid
from . import aaigrid
from . import raster

from .grid import RegularGrid, WarpedGrid
from .read import read_aai, read_gtiff, aairead, gtiffread
from .aaigrid import AAIGrid
from .raster import witch_of_agnesi, peaks, pad, slope, aspect, grad, div
from .raster import normed_vector_field

try:
    from .crfuncs import streamline2d
except ImportError:
    from .streamline import streamline2d

__all__ = ["grid", "aaigrid", "raster",
           "RegularGrid", "WarpedGrid",
           "aairead", "gtiffread", "read_aai", "read_gtiff",
           "AAIGrid",
           "pad", "slope", "aspect", "grad", "div", "normed_vector_field",
           "streamline2d"]




"""
Variogram estimation and modelling functions
"""

import numpy as np
try:
    from scipy.spatial.distance import pdist
    from scipy.optimize import minimize
except ImportError:
    pass
import random

class VariogramFunction(object):
    """ Base class for variogram functions, which can be composed
    through summation. """

    def __call__(self, lags):
        return self.func(lags)

    def __add__(self, other):
        return ComposedFunction(lambda lags: self(lags) + other(lags))

    def __rmul__(self, other):
        return ComposedFunction(lambda lags: other * self(lags))

class ComposedFunction(VariogramFunction):
    """ Type returned from function algebra """
    def __init__(self, f):
        self.func = lambda h: f(h)
        return

class Sph(VariogramFunction):
    """ Spherical model function """
    def __init__(self, rng):
        self.func = lambda h: np.where(h < rng, 1.5*h/rng - 0.5 * (h/rng)**3, 1.0)
        return

class Gau(VariogramFunction):
    """ Gaussian model function """
    def __init__(self, rng):
        self.func = lambda h: 1.0 - np.exp(-(h**2 / rng**2))
        return

class Nug(VariogramFunction):
    """ Nugget effect function """
    def __init__(self):
        self.func = lambda h: 1.0
        return

def fit_model(model, p, lags, variance):
    """ Fit a variogram model to an experimental variogram.

    Parameters:
    -----------

    model : function for creating variogram estimators that takes *lags* as an
            argument, e.g.

        model = lambda p: p[0] * Nug() +  p[1] * Sph(p[2])

    p : array of arguments for *model*
    lags : variogram offsets
    variance : variogram variance corresponding to lags
    """
    lags = np.array(lags)
    obj = lambda p: sum((model(p)(lags) - variance)**2)
    res = minimize(obj, p)
    return res.x

def estimate_vario(mp, npoints=2000, max_dist=None, interval=None):
    """ Given a Multipoint input, estimate a spatial variogram. By default, a
    variogram is quickly estimated. If npoints is None, then the full variogram
    is calculated.

    Parameters:
    -----------
    points : a Multipoint instance
    npoints : number of values use for variogram appromixation. If npoints is
              None, the full variogram is calculated.
    max_dist : the maximum lag
    interval : the lag interval

    Returns:
    --------
    lags : ndarray
    variogram : ndarray
    """
    if len(mp.data) != len(mp.vertices):
        raise Exception('estimate_variogram() requires a Multipoint with '
                        'scalar data')
    if npoints > len(mp):
        npoints = len(mp)

    irand = random.sample(np.arange(len(mp)), npoints)
    verts = mp.get_vertices()[irand]
    z = np.array([mp.data[i] for i in irand])

    if mp._crs == "lonlat":
        import warnings
        import pyproj
        warnings.warn("For improved performance, consider projecting data to an "
                      "approximately equidistant reference system first.")
        geod = pyproj.Geod(ellps="WGS84")
        def distfunc(a, b):
            return geod.inv(a[0], a[1], b[0], b[1])[2]
    else:
        distfunc = "euclidean"

    dist = pdist(verts, distfunc)
    I, J = ppairs(z)
    diff = z[I] - z[J]

    if max_dist is None:
        max_dist = dist.max()
    else:
        max_dist = float(max_dist)

    if interval is None:
        interval = max_dist / 10.0
    else:
        interval = float(interval)

    lags = np.arange(0, max_dist, interval)
    sigma_variance = np.empty_like(lags)
    for i, lag in enumerate(lags):
        band = (lag <= dist) * (dist < lag + interval)
        sigma_variance[i] = 0.5 * np.nanstd(diff[band])**2
    return lags, sigma_variance

def ppairs(A):
    """ For data *A*, return the pairwise differences.
    """
    n = len(A)
    fst = [i for i in range(n) for j in range(i+1,n)]
    scd = [j for i in range(n) for j in range(i+1,n)]
    return fst, scd




""" Classes for basic grid types """

import copy
import numbers
import numpy as np
from ..crs import Cartesian

IntegerType = (numbers.Integral, np.int32, np.int64)

class Grid(object):
    """ Grid base class. Intended only for implementing subclasses. The primary
    attributes defined by all Grid-derived classed are _hdr and data.
    """

    @property
    def crs(self):
        return getattr(self, "_crs", None)

    @property
    def nodata(self):
        return self._nodata

    def clipz(self, bounds):
        """ Clip the z-range in place to bounds = [min, max]. """
        self.values = self.values.clip(bounds[0], bounds[1])
        return

    def max(self):
        """ Return the maximum non-nan in self.data. """
        return np.nanmax(self.values)

    def min(self):
        """ Return the minimum non-nan in self.data. """
        return np.nanmin(self.values)

    def minmax(self):
        """ Return the minimum and maximum value of data array. """
        return (self.min(), self.max())

    def copy(self):
        """ Return a copy. """
        return copy.deepcopy(self)


class RegularGrid(Grid):
    """ Regular (structured) grid class. A RegularGrid contains a fixed number
    of rows and columns with a constant spacing and a scalar or vector field
    defined as `values`.

    Positions on a RegularGrid are referenced to their array indices *(i,j)*
    using an affine transform *T* such that

    (a, b, c, d, e, f) = T

    X = a + j * c + i * e
    Y = b + i * d + j * f

    in which *(a,b)* defines the grid origin, *(c,d)* defines the resolution in
    the horizontal and vertical directions, and *(e,f)* can be used to define a
    rotation. In the common case of an orthogonal grid with north "up",

    e = f = 0
    """
    def __init__(self, transform, values=None, crs=None, nodata_value=None):
        """ Create a RegularGrid instance with cells referenced according to
        *transform*, which is an iterable or dictionary consisting of
        (xllcenter, yllcenter, dx, dy, xrot, yrot).

        Optional parameters:
        --------------------
        values : grid data (nrows x ncols) [default NaN]
        crs : Grid coordinate reference system [default None]
        """
        if hasattr(transform, "keys"):
            self._transform = tuple([transform[f] for f in
                    ("xllcenter", "yllcenter", "dx", "dy", "xrot", "yrot")])
        elif len(transform) == 6:
            self._transform = tuple(transform)
        elif len(transform) == 4:
            self._transform = (transform[0], transform[1],
                               transform[2], transform[3], 0, 0)
        else:
            raise GridError("RegularGrid must be initialized with a transform"
                            "iterable or dictionary")

        if values is not None:
            self.values = values
        else:
            self.values = np.atleast_2d([np.nan])

        if crs is None:
            self._crs = Cartesian
        else:
            self._crs = crs

        if nodata_value is None:
            self._nodata = get_nodata(self.values.dtype.type)
        else:
            self._nodata = nodata_value
        return

    def __add__(self, other):
        if self._equivalent_structure(other):
            return RegularGrid(copy.copy(self.transform),
                               values=self.values+other.values)
        else:
            raise NonEquivalentGridError(self, other)

    def __sub__(self, other):
        if self._equivalent_structure(other):
            return RegularGrid(copy.copy(self.transform),
                               values=self.values-other.values)
        else:
            raise NonEquivalentGridError(self, other)

    def _equivalent_structure(self, other):
        return (self._transform == other._transform) and \
               (self.values.shape == other.values.shape)

    @property
    def transform(self):
        return self._transform

    @property
    def size(self):
        return self.values.shape

    def center_llref(self):
        """ Return the 'lower-left' reference in terms of a center coordinate.
        """
        return self._transform[0], self._transform[1]

    def corner_llref(self):
        """ Return the 'lower-left' reference in terms of a center coordinate.
        """
        xllcorner = self._transform[0] - 0.5 * self._transform[2]
        yllcorner = self._transform[1] - 0.5 * self._transform[3]
        return (xllcorner, yllcorner)

    def coordmesh(self, *args, **kwargs):
        """ Alias for center_coords() """
        return self.center_coords(*args, **kwargs)

    def center_coords(self):
        t = self._transform
        xcoords = np.empty(self.values.shape[:2])
        ycoords = np.empty(self.values.shape[:2])
        irow = np.arange(self.values.shape[0])
        for i in range(self.values.shape[1]):
            xcoords[:,i] = t[0] + i*t[2] + irow*t[4]
            ycoords[:,i] = (t[1] + irow*t[3] + i*t[5])
        return xcoords, ycoords

    def vertex_coords(self):
        """ Return the coordinates of cell vertices. """
        xllcorner, yllcorner = self.corner_llref()
        t = self._transform
        ny, nx = self.values.shape[:2]
        xurcorner = xllcorner + (nx+1) * t[2] + (ny+1) * t[4]
        yurcorner = yllcorner + (ny+1) * t[3] + (nx+1) * t[5]
        return np.meshgrid(np.linspace(xllcorner, xurcorner, nx + 1),
                           np.linspace(yurcorner, yllcorner, ny + 1))

    @property
    def bbox(self):
        extents = self.get_extents(reference="edge")
        return extents[0], extents[2], extents[1], extents[3]

    def get_extents(self, reference='center'):
        """ Return the region characteristics as a tuple (xmin, xmax, ymin,
        ymax). *reference* is a string and may be 'center' or 'edge'. """
        if reference == 'center':
            x0, y0 = self.center_llref()
            n = -1
        elif reference == 'edge':
            x0, y0 = self.corner_llref()
            n = 0
        else:
            raise GridError("`reference` must be 'center' or 'edge'")
        ny, nx = self.values.shape[:2]
        dx, dy = self._transform[2:4]
        sx, sy = self._transform[4:]

        sgn = lambda a: 0 if a==0 else a/abs(a)
        if sgn(dx) == sgn(sx):
            xrng = dx*(nx+n) + sx*(ny+n)
        else:
            xrng = dx*(nx+n) - sx*(ny+n)

        if sgn(dy) == sgn(sy):
            yrng = dy*(ny+n) + sy*(nx+n)
        else:
            yrng = dy*(ny+n) - sy*(nx+n)
        return min(x0, x0+xrng), max(x0, x0+xrng), min(y0, y0+yrng), max(y0, y0+yrng)

    def clip(self, xmin, xmax, ymin, ymax, crs=None):
        """ Return a clipped version of grid constrained to a bounding box
        defined by *xmin*, *xmax*, *ymin*, *ymax*. If *crs* is not provided, it
        is assumed that the bounding box shares the same coordinate system as
        the grid. """
        if crs is not None:
            xg, yg = crs.project([xmin, xmax], [ymin, ymax], inverse=True)
            x, y = self.crs.project(xg, yg)
            xmin, xmax = x
            ymin, ymax = y
        else:
            crs = self.crs

        ll = self.get_indices(xmin, ymin)
        lr = self.get_indices(xmax, ymin)
        ul = self.get_indices(xmin, ymax)
        ur = self.get_indices(xmax, ymax)

        i0 = min(ll[0], lr[0], ul[0], ur[0])
        i1 = max(ll[0], lr[0], ul[0], ur[0]) + 1
        j0 = min(ll[1], lr[1], ul[1], ur[1])
        j1 = max(ll[1], lr[1], ul[1], ur[1]) + 1

        values = self.values[i0:i1,j0:j1].copy()
        t = self.transform
        x0 = t[0] + j0*t[2] + i0*t[4]
        y0 = t[1] + i0*t[3] + j0*t[5]
        tnew = (x0, y0, t[2], t[3], t[4], t[5])
        return RegularGrid(tnew, values, crs=self.crs)

    def resample_griddata(self, dx, dy, method='nearest'):
        """ Resample array to have spacing `dx`, `dy' using *scipy.griddata*

        Warning: This method is likely to be removed in the future, once
        the built-in `resample` method matures.

        Parameters
        ----------

        dx : cell dimension, float

        dy : cell dimension, float

        method : interpolation method, string ('nearest', 'linear')
        """
        from scipy.interpolate import griddata
        ny0, nx0 = self.values.shape[:2]
        dx0, dy0 = self._transform[2:4]
        xllcenter, yllcenter = self.center_llref()
        xurcenter = xllcenter + dx0*nx0
        yurcenter = yllcenter + dy0*ny0
        nx = int((xurcenter - xllcenter) // dx)
        ny = int((yurcenter - yllcenter) // dy)

        xx0, yy0 = self.coordmesh()
        xx, yy = np.meshgrid(np.linspace(xllcenter, xurcenter, nx),
                             np.linspace(yllcenter, yurcenter, ny))
        idata = griddata((xx0.flatten(), yy0.flatten()), self.values.flatten(),
                         (xx.flatten(), yy.flatten()), method=method)
        values = idata.reshape(ny, nx)
        t = self._transform
        tnew = (t[0], t[1], dx, dy, t[4], t[5])
        return RegularGrid(tnew, values)

    def resample(self, dx, dy, method='nearest'):
        """ Resample array to have spacing `dx`, `dy'.

        Parameters
        ----------

        dx : cell dimension, float

        dy : cell dimension, float

        method : interpolation method, string ('nearest',)
        """
        ny0, nx0 = self.values.shape[:2]
        dx0, dy0 = self._transform[2:4]
        xllcenter, yllcenter = self.center_llref()

        if method == 'nearest':
            rx, ry = dx / dx0, dy / dy0
            I = np.around(np.arange(ry/2, self.values.shape[0], ry)).astype(int)
            J = np.around(np.arange(rx/2, self.values.shape[1], rx)).astype(int)
            JJ, II = np.meshgrid(J, I)
            values = self.values[II, JJ]
        else:
            raise NotImplementedError('method "{0}" not '
                                      'implemented'.format(method))

        t = self._transform
        tnew = (t[0], t[1], dx, dy, t[4], t[5])
        return RegularGrid(tnew, values)

    def get_positions(self, x, y):
        """ Return the column and row indices for the point nearest
        geographical coordinates (x, y). """
        # Calculate this by forming block matrices
        #       | dx sy          |
        #       | sx dy          |
        #   T = |      ....      |
        #       |          dx sy |
        #       |          sx dy |
        #
        #       | x0 |
        #       | y0 |
        #   S = | .. |
        #       | x0 |
        #       | y0 |
        #
        # where the grid transform is t = (x0, y0, dx, dy, sy, sx)
        #
        # Then the nearest indices J come from solving the system
        # T J = X - S
        try:
            npts = len(x)
        except TypeError:
            npts = 1

        n = min(70, npts)    # number to process at once
        t = self._transform
        T = (np.diag(np.tile([t[2], t[3]], n)) +
             np.diag(np.tile([t[4], 0], n)[:-1], 1) +
             np.diag(np.tile([t[5], 0], n)[:-1], -1))
        S = np.tile([t[0], t[1]], n)

        if npts == 1:
            ind = np.linalg.solve(T, np.array([x, y]) - S)
        else:
            ind = np.empty(2*npts, dtype=np.float64)
            i = 0
            while i != npts:
                ip = min(i + n, npts)
                xy = np.zeros(2*n, dtype=np.float64)
                xy[:2*(ip-i)] = np.vstack([x[i:ip], y[i:ip]]).T.ravel()
                ind_ = np.linalg.solve(T, xy - S)
                ind[2*i:2*ip] = ind_[:2*(ip-i)]
                i = ip

        ny, nx = self.size
        j = ind[::2]
        i = ind[1::2]
        return i, j

    def get_indices(self, x, y):
        """ Return the column and row indices for the point nearest
        geographical coordinates (x, y). """
        ny, nx = self.size
        i, j = self.get_positions(x, y)

        if len(i) != 1:
            i, j = np.round(i).astype(int), np.round(j).astype(int)
            if min(i) < 0 or max(i) > ny-1 or min(j) < 0 or max(j) > nx-1:
                raise GridError("Coordinates outside grid region ({0})".format(self.bbox))
        else:
            i, j = int(round(i[0])), int(round(j[0]))
            if not (0 <= i <= ny-1) or not(0 <= j <= nx-1):
                raise GridError("Coordinates outside grid region ({0})".format(self.bbox))

        return i,j

    def sample_nearest(self, x, y):
        """ Return the value nearest to (`x`, `y`). Nearest grid center
        sampling scheme. """
        i, j = self.get_indices(x, y)
        ny, nx = self.values.shape[:2]
        return self.values[i, j]

    def sample_bilinear(self, x, y):
        """ Return the value nearest to (`x`, `y`). Bilinear sampling scheme.
        """
        i, j = self.get_positions(x, y)
        i0 = np.floor(i).astype(int)
        i1 = np.ceil(i).astype(int)
        j0 = np.floor(j).astype(int)
        j1 = np.ceil(j).astype(int)

        # Handle case where interpolation point is exactly on a grid point
        mska = i0==i1
        mskb = i0==0
        i0[mska&~mskb] -= 1
        i1[mska&mskb] += 1

        mska = j0==j1
        mskb = j0==0
        j0[mska&~mskb] -= 1
        j1[mska&mskb] += 1

        dx, dy = self._transform[2:4]
        z = (self.values[i0,j0]*(i1-i)*(j1-j) + self.values[i1,j0]*(i-i0)*(j1-j) + \
             self.values[i0,j1]*(i1-i)*(j-j0) + self.values[i1,j1]*(i-i0)*(j-j0))
        return z

    def sample(self, x, y, crs=None, method="bilinear"):
        """ Return the values nearest (*x*, *y*), where *x* and *y* may be
        equal length vectors. Keyword *method* may be one of 'nearest',
        'bilinear' (default).
        """
        if crs is not None:
            xg, yg = crs.project(x, y, inverse=True)
            x, y = self.crs.project(xg, yg)

        if method == "nearest":
            return self.sample_nearest(x, y)
        elif method == "bilinear":
            return self.sample_bilinear(x, y)
        else:
            raise ValueError("method \"{0}\" not available".format(method))

    def profile(self, line, resolution=None, **kw):
        """ Sample along a *line* at *resolution*.

        Parameters:
        -----------
        line : `geometry.Line`-like object describing the sampling path

        resolution : sample spacing

        Additional keyword arguments passed to `RegularGrid.sample` (e.g. to
        specify sampling method)

        Returns:
        --------
        profile : ndarray
        """
        if resolution is None:
            resolution = min(self.transform[2:4])

        remainder = 0
        pt0 = line[0]
        vertices = [pt0.get_vertex(self.crs)]

        for seg in line.segments:
            pos = 0
            az = seg[0].azimuth(seg[1])

            while pos < seg.length:
                distance_to_endpt = pt0.distance(seg[1])
                if distance_to_endpt >= resolution:
                    pt1 = pt0.walk(resolution - remainder, az)
                    pos += resolution - remainder
                    vertices.append(pt1.get_vertex(self.crs))
                    remainder = 0
                    pt0 = pt1
                else:
                    remainder = distance_to_endpt
                    pos = seg.length
                    pt0 = seg[1]

        z = self.sample(*zip(*vertices), **kw)
        return vertices, z

    def as_warpedgrid(self):
        """ Return a copy of grid as a `WarpedGrid`. This is a more general
        grid class that has a larger memory footprint but can represent more
        flexible data layouts.
        """
        Xc, Yc = self.center_coords()
        return WarpedGrid(Xc, Yc, self.values.copy())

    def gtiffwrite(self, fnm):
        """ Write data to a GeoTiff file using GDAL """
        try:
            from . import _gtiff
            return _gtiff.write(fnm, self)
        except ImportError as e:
            raise ImportError("{0}\nReading GeoTiffs depends on GDAL, which "
                              "could not be imported".format(e))

    def aaiwrite(self, f, reference='corner', nodata_value=-9999):
        """ Save internal data as an ASCII grid. Based on the ESRI standard,
        only isometric grids (i.e. `hdr['dx'] == hdr['dy']` can be saved,
        otherwise `GridIOError` is thrown.

        Parameters:
        -----------
        f : a file-like object or a filename

        reference : specify a header reference ('center' | 'corner')

        nodata_value : specify how NaNs should be represented (int or float)
        """
        if reference not in ('center', 'corner'):
            raise GridIOError("reference in AAIGrid.tofile() must be 'center' or "
                           "'corner'")

        if np.any(self._transform[4:] != 0.0):
            raise GridIOError("ESRI ASCII grids do not support skewed grids")

        ny, nx = self.values.shape[:2]
        x0, y0, dx, dy = self._transform[:4]
        if dx != dy:
            raise GridIOError("ASCII grids require isometric grid cells")

        if not hasattr(f, 'read'):
            f = open(f, "w")

        try:
            data_a = self.values.copy()
            data_a[np.isnan(data_a)] = nodata_value

            f.write("NCOLS {0}\n".format(nx))
            f.write("NROWS {0}\n".format(ny))
            if reference == 'center':
                f.write("XLLCENTER {0}\n".format(x0))
                f.write("YLLCENTER {0}\n".format(y0))
            elif reference == 'corner':
                xllcorner, yllcorner = self.corner_llref()
                f.write("XLLCORNER {0}\n".format(xllcorner))
                f.write("YLLCORNER {0}\n".format(yllcorner))
            f.write("CELLSIZE {0}\n".format(dx))
            f.write("NODATA_VALUE {0}\n".format(nodata_value))
            f.writelines([str(row).replace(',','')[1:-1] +
                            '\n' for row in data_a.tolist()])
        finally:
            f.close()
        return

class WarpedGrid(Grid):
    """ Warped Grid class. A WarpedGrid contains a fixed number of rows
    and columns and a scalar or vector field defined on the z-axis. Grid
    spacing is not necessarily constant.
    """

    def __init__(self, X, Y, values, nodata_value=None):
        """
        Parameters:
        -----------
        X : first-dimension coordinates of grid centers

        Y : second-dimension coordinates of grid centers

        values : dependent m-dimensional quantity (nrows x ncols)
        """

        if any(a is None for a in (X, Y, values)):
            raise GridError('All of (X, Y, values) must be provided')

        if not (X.shape == Y.shape == values.shape[:2]):
            raise GridError('All of (X, Y, values) must share the same '
                            'size over the first two dimensions')

        self.X = X
        self.Y = Y
        self.values = values

        if nodata_value is None:
            self._nodata = get_nodata(self.values.dtype.type)
        else:
            self._nodata = nodata_value
        return

    def __add__(self, other):
        if self._equivalent_structure(other):
            return WarpedGrid(self.X.copy(), self.Y.copy(), self.values+other.values)
        else:
            raise NonEquivalentGridError(self, other)

    def __sub__(self, other):
        if self._equivalent_structure(other):
            return WarpedGrid(self.X.copy(), self.Y.copy(), self.values-other.values)
        else:
            raise NonEquivalentGridError(self, other)

    def _equivalent_structure(self, other):
        return np.all(self.X == other.X) and np.all(self.Y == other.Y) and \
                np.all(self.values.shape == other.values.shape)

    def rotate(self, deg, origin=(0.0, 0.0)):
        """ Rotate grid by *deg* degrees counter-clockwise around *origin*. """
        raise NotImplementedError

    def resample(self, X, Y):
        """ Resample internal grid to the points defined by *X*, *Y*. """
        raise NotImplementedError


class GridError(Exception):
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message


class GridIOError(GridError):
    def __init__(self, message=''):
        self.message = message


class NonEquivalentGridError(GridError):
    def __init__(self, A, B, message=''):
        if len(message) == 0:
            self.message = ("{0} and {1} do not share equivalent "
                            "grid layouts".format(A, B))
        else:
            self.message = message

def get_nodata(T):
    """ Return a default value for NODATA given a type (e.g. int, float,
    complex).
    """
    if issubclass(T, IntegerType):
        return -9999
    elif issubclass(T, (numbers.Real, numbers.Complex)):
        return np.nan
    else:
        raise ValueError("No default NODATA value for type {0}".format(T))




"""
Python ESRI ASCII grid driver

The primary class is AAIGrid, which can be called with a filename
argument to open a grid, or can read data from a numpy array.

Written by Nat Wilson
"""

from math import sqrt
import numpy as np
from . import grid
import traceback

class AAIGrid(grid.RegularGrid):
    """ Grid class designed to read and emit ESRI ASCII grid files. """
    _hdr = {}
    aschdr = {}
    aschdr['ncols'] = None
    aschdr['nrows'] = None
    aschdr['yllcenter'] = None
    aschdr['xllcenter'] = None
    aschdr['yllcorner'] = None
    aschdr['xllcorner'] = None
    aschdr['cellsize'] = None
    aschdr['nodata_value'] = None

    def __init__(self, incoming=None, hdr=None):
        """ Contains the table of data from an ESRI ASCII raster file.

        Parameters
        ----------
        incoming : either a filename or an ndarray containing the
                   raster data. If incoming is not provided, then
                   read() must be called later with a suitable
                   filename.
        hdr : dictionary containing AAIGrid header information

        Header dictionaries contain the following keys:
            ncols           (int)
            nrows           (int)
            either:
                xllcenter   (int)
                yllcenter   (int)
            or:
                xllcorner   (int)
                yllcorner   (int)
            cellsize        (int)
            nodata_value    (int)
        """

        if hdr is not None:
            self.aschdr = hdr

        if incoming is not None:
            try:
                # assume incoming is a file
                self.fromfile(incoming)
            except (TypeError, UnicodeDecodeError):
                # isn't a file, so hopefully an ndarray
                self.fromarray(incoming, hdr)
        elif hdr is not None:
            self.data = np.empty((hdr['nrows'], hdr['ncols']))
        else:
            self.data = None

        self._check_header()
        self._enforce_hdr_consistency()
        return

    def __str__(self):
        if self.data is None:
            stat = 'empty'
        else:
            stat = 'populated'
        return "{status} AAIGrid instance".format(status=stat)

    def __len__(self):
        if self.data is None:
            return 0
        else:
            return len(self.data)

    def __add__(self, other):
        if isinstance(other, AAIGrid):
            if (self.data is not None) and (other.data is not None):
                return AAIGrid(self.data + other.data, self.aschdr)
            else:
                raise ValueError("cannot add NoneType data array")
        else:
            try:
                return AAIGrid(self.data + other, self.aschdr)
            except AttributeError:
                raise AAIError("self.data not defined")
            except Exception:
                raise AAIError("addition with type {0} not "
                               "defined".format(type(other)))

    def __sub__(self, other):
        if isinstance(other, AAIGrid):
            if (self.data is not None) and (other.data is not None):
                return AAIGrid(self.data - other.data, self.aschdr)
            else:
                raise ValueError("cannot subtract NoneType data array")
        else:
            try:
                return AAIGrid(self.data - other, self.aschdr)
            except AttributeError:
                raise AAIError("self.data not defined")
            except Exception:
                raise AAIError("subtraction with type {0} not "
                               "defined".format(type(other)))

    def __mul__(self, other):
        if isinstance(other, AAIGrid):
            if (self.data is not None) and (other.data is not None):
                return AAIGrid(self.data * other.data, self.aschdr)
            else:
                raise ValueError("cannot multiply NoneType data array")
        else:
            try:
                return AAIGrid(self.data * other, self.aschdr)
            except AttributeError:
                raise AAIError("self.data not defined")
            except:
                raise AAIError("multiplication with type{0} not "
                                "defined".format(type(other)))

    def __div__(self, other):
        if isinstance(other, AAIGrid):
            if (self.data is not None) and (other.data is not None):
                return AAIGrid(self.data / other.data, self.aschdr)
            else:
                raise ValueError("cannot divide NoneType data array")
        else:
            try:
                return AAIGrid(self.data / other, self.aschdr)
            except AttributeError:
                raise AAIError("self.data not defined")
            except:
                raise AAIError("division with type{0} not "
                                "defined".format(type(other)))

    def _check_header(self, hdr=None):
        """ Make sure that all required header records are present. """

        if hdr is None:
            hdr = self.aschdr

        for field in ('ncols', 'nrows', 'cellsize'):
            if hdr.get(field) is None:
                raise AAIError("{0} not defined".format(field.upper()))

        for field in ('ncols', 'nrows'):
            hdr[field] = int(hdr[field])

        if hdr.get('nodata_value') is None:
            hdr['nodata_value'] = -9999

        hdr = self._check_header_references(hdr)

        return hdr

    def _check_header_references(self, hdr):
        """ Make sure that both corner and center reference coordinates are
        present. """
        d = hdr['cellsize']
        if hdr.get('yllcenter') is None:
            if hdr.get('yllcorner') is None:
                raise AAIError('y reference not defined')
            else:
                hdr['yllcenter'] = hdr['yllcorner'] + d / 2.0
        else:
            hdr['yllcorner'] = hdr['yllcenter'] - d / 2.0

        if hdr.get('xllcenter') is None:
            if hdr.get('xllcorner') is None:
                raise AAIError('x reference not defined')
            else:
                hdr['xllcenter'] = hdr['xllcorner'] + d / 2.0
        else:
            hdr['xllcorner'] = hdr['xllcenter'] - d / 2.0
        return hdr

    def _enforce_hdr_consistency(self):
        """ Make sure that base `hdr` remains consistent with AAIGrid-specific
        `aschdr`. """
        self._hdr['dx'] = self.aschdr['cellsize']
        self._hdr['dy'] = self.aschdr['cellsize']
        self._hdr['nx'] = self.aschdr['ncols']
        self._hdr['ny'] = self.aschdr['nrows']
        self._hdr['xllcorner'] = self.aschdr['xllcorner']
        self._hdr['yllcorner'] = self.aschdr['yllcorner']
        return

    def get_indices(self, x, y):
        """ Return the column and row indices for the point nearest
        geographical coordinates (x, y). """
        if self.data is None:
            raise AAIError('no raster to query')

        self.aschdr = self._check_header_references(self.aschdr)

        x0 = self.aschdr['xllcenter']
        y0 = self.aschdr['yllcenter']
        d = self.aschdr['cellsize']
        nx, ny = self.data.shape

        xi = np.clip(
                (np.around((np.array(x) - x0) / d)).astype(int),
                0, nx-1)
        yi = np.clip(
                self.data.shape[0] -
                    (np.around((np.array(y)-y0) / d)).astype(int) - 1,
                0, ny-1)
        return xi, yi

    def get_region(self, reference='center'):
        """ Return the region characteristics as a tuple, relative to either
        cell centers or the grid edges. """
        if reference == 'center':
            x0, y0  = self.aschdr['xllcenter'], self.aschdr['yllcenter']
            n = 0
        elif reference == 'edge':
            x0, y0  = self.aschdr['xllcenter'], self.aschdr['yllcenter']
            n = 1
        else:
            raise AAIError("`reference` must be 'center' or 'edge'")
        return (x0, x0 + (self.aschdr['ncols'] + n) * self.aschdr['cellsize'],
                y0, y0 + (self.aschdr['nrows'] + n) * self.aschdr['cellsize'])

    def coordmesh(self, grid_registration='center'):
        """ Return a pair of arrays containing the *X* and *Y* coordinates of
        the grid. """
        if grid_registration == 'center':
            xll = self.aschdr['xllcenter']
            yll = self.aschdr['yllcenter']
        elif grid_registration == 'corner':
            xll = self.aschdr['xllcorner']
            yll = self.aschdr['yllcorner']
        else:
            raise AAIError("grid_registration must be 'center' or 'corner'\n")

        X = (xll + np.arange(self.data.shape[1]) * self.aschdr['cellsize'])
        Y = (yll + np.arange(self.data.shape[0])[::-1] * self.aschdr['cellsize'])
        return np.meshgrid(X, Y)

    def max(self):
        """ Return the maximum non-nan in self.data. """
        return self.data[np.isnan(self.data)==False].max()

    def min(self):
        """ Return the minimum non-nan in self.data. """
        return self.data[np.isnan(self.data)==False].min()

    def read(self, fnm):
        """ Read an existing ASCII grid file. """
        return self.fromfile(fnm)

    def fromfile(self, fnm):
        """ Read an existing ASCII grid file. """
        try:
            h = []
            with open(fnm, 'r') as f:
                # Count the number of header entries
                cnt = 0
                while True:
                    l = f.readline()
                    if l.split(None, 1)[0].lower() in ['nrows', 'ncols',
                            'yllcenter', 'xllcenter', 'yllcorner', 'xllcorner',
                            'cellsize', 'nodata_value']:
                        cnt += 1
                    else:
                        break

                # Read the header, then the array data
                f.seek(0)
                for _ in range(cnt):
                    h.append(f.readline())
                data = f.readlines()

        except IOError:
            raise AAIError('error while trying to open {0}'.format(fnm))

        # Store data internally
        try:
            hs = [rec.split(None, 1) for rec in h]
            hdr = dict([(rec[0].lower(), float(rec[1])) for rec in hs])
            if 'yllcenter' not in hdr.keys():
                hdr['yllcenter'] = None

            if 'xllcenter' not in hdr.keys():
                hdr['xllcenter'] = None

            if 'yllcorner' not in hdr.keys():
                hdr['yllcorner'] = None

            if 'xllcorner' not in hdr.keys():
                hdr['xllcorner'] = None

            if 'nodata_value' not in hdr.keys():
                hdr['nodata_value'] = -9999

        except:
            traceback.print_exc()
            return

        try:
            hdr = self._check_header(hdr)
        except AAIError as s:
            print(s)
            return

        self.aschdr = hdr

        try:
            f = lambda l: [float(i) for i in l.split()]
            data_f = map(f, data)
            data_a = np.array(data_f)
            data_a[data_a==hdr['nodata_value']] = np.nan
            self.data = data_a
        except:
            traceback.print_exc()

        return

    def fromarray(self, A, hdr):
        """ Read data from a 2d numpy array. The hdr argument is a dictionary
        of header information.

        ncols : int
            Number of columns

        nrows : int
            Number of rows

        cellsize : float
            Cell dimension

        nodata_value : float
            Value for missing data

        xllcenter OR xllcorner : float
            Lower left cell *x*-registration

        yllcenter OR yllcorner : float
            Lower left cell *y*-registration
        """
        # Should add sanity checks
        self.aschdr = hdr
        #A[np.isnan(A)] = hdr['nodata_value']
        self.data = A.copy()[:,:]

    def tofile(self, f, reference='center'):
        """ Save internal data to *f*, either a file-like object or a filename.

        Parameters
        ----------
        f : filename or file handle
        reference : grid registration ("center" or "corner")
        """
        if self.data is None:
            raise AAIError("no data to write!")

        if reference not in ('center', 'corner'):
            raise AAIError("reference in AAIGrid.tofile() must be 'center' or "
                           "'corner'")

        try:
            f.read()
        except AttributeError:
            f = open(f, "w")
        except IOError:
            pass

        try:
            data_a = self.data.copy()
            data_a[np.isnan(data_a)] = self.aschdr['nodata_value']

            f.write("NCOLS {0}\n".format(self.aschdr['ncols']))
            f.write("NROWS {0}\n".format(self.aschdr['nrows']))
            if reference == 'center':
                f.write("XLLCENTER {0}\n".format(self.aschdr['xllcenter']))
                f.write("YLLCENTER {0}\n".format(self.aschdr['yllcenter']))
            elif reference == 'corner':
                f.write("XLLCORNER {0}\n".format(self.aschdr['xllcorner']))
                f.write("YLLCORNER {0}\n".format(self.aschdr['yllcorner']))
            f.write("CELLSIZE {0}\n".format(self.aschdr['cellsize']))
            f.write("NODATA_VALUE {0}\n".format(self.aschdr['nodata_value']))
            f.writelines([str(row).replace(',','')[1:-1] +
                            '\n' for row in data_a.tolist()])

        except:
            traceback.print_exc()
        finally:
            f.close()
        return

    def toarray(self):
        """ Return a 2d array with internal data. Header information can be
        obtained with AAIGrid.hdr.
        """
        return self.data.copy()[::-1,:]

    def minmax(self):
        """ Return the minimum and maximum value of data array. """
        A = self.data[np.isnan(self.data)==False]
        return (A.min(), A.max())

    def sample(self, x, y):
        """ Return the value nearest to *x*, *y*. """
        xi, yi = self.get_indices(x, y)
        ncols = self.aschdr['ncols']
        nrows = self.aschdr['nrows']
        if (np.any(xi < 0) or np.any(xi >= ncols) or
            np.any(yi < 0) or np.any(yi >= nrows)):
            raise AAIError("coordinates are outside grid region",
                detail="({0}, {1}), ({2}, {3})".format(xi, yi, ncols, nrows))
        else:
            z = self.data[yi, xi]
        return z

    def get_profile(self, segments, resolution=10.0):
        """ Sample along a line defined as `segments`. Does not interpolate.

        Parameters
        ----------
        segments : iterable containing (x,y) pairs
        resolution : sample spacing

        Returns
        -------
        profile : ndarray
        """
        z = []
        p = 0
        for s, f in zip(segments[:-1], segments[1:]):

            x0 = s[0]
            y0 = s[1]
            xf = f[0]
            yf = f[1]

            xlen = xf-x0
            ylen = yf-y0
            d = sqrt((xf - x0)**2 + (yf - y0)**2)

            while p < d:
                fd = p / d
                fx = fd*xlen
                fy = fd*ylen
                xi, yi = self.get_indices(x0+fx, y0+fy)
                z.append(self.data[yi, xi])
                p += resolution
            p -= d

        return np.array(z)

    def clip(self, bounds):
        """ Clip the z-range in place to bounds = [min, max]. """
        self.data = self.data.clip(bounds[0], bounds[1])
        return

    def resample(self, cellsize, method='nearest'):
        """ Resample array to have `cellsize`. Modify header in-place.

        Parameters
        ----------
        cellsize : cell dimension, float
        method : interpolation method, string ('nearest', 'linear')
        """
        xllcenter = self.aschdr['xllcenter']
        yllcenter = self.aschdr['yllcenter']
        xurcenter = xllcenter + self.aschdr['cellsize'] * self.aschdr['ncols']
        yurcenter = yllcenter + self.aschdr['cellsize'] * self.aschdr['nrows']
        nx = int((xurcenter - xllcenter) // cellsize)
        ny = int((yurcenter - yllcenter) // cellsize)
        dimratio = cellsize / self.aschdr['cellsize']

        if method == 'nearest':
            JJ, II = np.meshgrid(np.arange(nx), np.arange(ny))
            srcII = np.around(II * dimratio) \
                            .astype(int) \
                            .clip(0, self.aschdr['nrows'] - 1)
            srcJJ = np.around(JJ * dimratio) \
                            .astype(int).\
                            clip(0, self.aschdr['ncols'] - 1)
            self.data = self.data[srcII, srcJJ]
        else:
            raise NotImplementedError('method "{0}" not '
                                      'implemented'.format(method))

        self.aschdr['cellsize'] = float(cellsize)
        self.aschdr['nrows'] = ny
        self.aschdr['ncols'] = nx
        self.aschdr['xllcorner'] = self.aschdr['xllcenter'] - (0.5 * cellsize)
        self.aschdr['yllcorner'] = self.aschdr['yllcenter'] - (0.5 * cellsize)
        self._enforce_hdr_consistency()
        return

    def resize(self, te):
        """ Resize array to fit within extents given by te. If the new
        dimensions are smaller, the data is clipped. If they are larger,
        nan padding is added.

        Parameters
        ----------
        te : tuple of center coordinates in the form (xmin, xmax, ymin, ymax).

        Returns None.
        """
        if self.data is not None:
            xmin1 = self.aschdr['xllcenter']
            xmax1 = xmin1 + self.aschdr['cellsize'] * (self.aschdr['ncols'] - 1)
            ymin1 = self.aschdr['yllcenter']
            ymax1 = ymin1 + self.aschdr['cellsize'] * (self.aschdr['nrows'] - 1)

            xmin2 = te[0]
            xmax2 = te[1]
            ymin2 = te[2]
            ymax2 = te[3]

            data_a = self.data.copy()
            ny, nx = data_a.shape

            # The left side
            dx = int(np.floor((xmin2-xmin1) / float(self.aschdr['cellsize'])))
            if dx < 0:
                data_a = np.hstack([np.nan*np.ones([ny, dx]), data_a])
            elif dx > 0:
                data_a = data_a[:,dx:]
            self.aschdr['xllcenter'] = xmin1 + self.aschdr['cellsize'] * dx

            # The right side
            dx = int(np.floor((xmax2-xmax1) / float(self.aschdr['cellsize'])))
            if dx > 0:
                data_a = np.hstack([data_a, np.nan*np.ones([ny, dx])])
            elif dx < 0:
                data_a = data_a[:,:dx]
            self.aschdr['ncols'] = data_a.shape[1]

            ny, nx = data_a.shape

            # The bottom
            dy = int(np.ceil((ymin2-ymin1) / float(self.aschdr['cellsize'])))
            if dy < 0:
                data_a = np.vstack([data_a, np.nan*np.ones([dy, nx])])
            elif dy > 0:
                data_a = data_a[dy:,:]
            self.aschdr['yllcenter'] = ymin1 + self.aschdr['cellsize'] * dy

            # The top
            dy = int(np.floor((ymax2-ymax1) / float(self.aschdr['cellsize'])))
            if dy > 0:
                data_a = np.vstack([np.nan*np.ones([dy, nx]), data_a])
            elif dy < 0:
                data_a = data_a[:dy,:]
            self.aschdr['nrows'] = data_a.shape[0]

            self.data = data_a
            self._check_header()
            self._enforce_hdr_consistency()

        else:
            raise AAIError("no data to resize")
        return

class AAIError(Exception):
    """ Exceptions related to ESRI ASCII grid driver. """
    def __init__(self, value, detail=None):
        self.value = value
        self.detail = detail
    def __str__(self):
        if self.detail is not None:
            s = self.value + ': ' + self.detail
        else:
            s = self.value
        return repr(s)




""" Vector geospatial metadata management """

from copy import deepcopy
import numbers
from collections import Mapping
from numpy import isnan, int32, int64
IntegerType = (numbers.Integral, int32, int64)

class Metadata(Mapping):
    """ Class for handling collections of metadata. Data are organized similar
    to a Python dict, however different values must all have the same length
    and be of uniform type.

    Data are accessed either by field name (returning the entire series for
    that field) for by index (returning the value within eather field at that
    index). """

    _data = {}
    _fieldtypes = []

    def __init__(self, data, singleton=False, copy=False):
        """ Create a collection of metadata from *data*, which may be a list
        with uniform type or a dictionary with equally-sized fields of uniform
        type.

        Parameters
        ----------
        data : list of uniform type or dictionary of equally-sized fields of
        uniform type

        *kwargs*:
        singleton : treat data as a single unit, rather than as a list of units
        """
        if copy:
            data = deepcopy(data)

        if data is None:
            self._len = 0
            self._data = {}
            self._fieldtypes = []

        elif singleton:

            if hasattr(data, "keys") and hasattr(data.values, "__call__"):
                self._data = data
            else:
                self._data = {"values": data}
            self._fieldtypes = [type(self._data[k]) for k in self._data]
            self._len = 1

        else:

            if hasattr(data, 'keys') and hasattr(data.values, '__call__'):
                # Dictionary of attributes
                n = -1
                for k in data:
                    dtype = type(data[k][0])
                    if not all(isinstance(a, dtype) or a is None or isnan(a)
                                    for a in data[k]):
                        raise MetadataError("Data must have uniform type")
                    if n != -1:
                        if len(data[k]) != n:
                            raise MetadataError("Data must have uniform lengths")
                    else:
                        n = len(data[k])

                self._len = n
                self._data = data

            else:
                # Single attribute
                if not hasattr(data, '__iter__'):
                    data = [data]
                self._len = len(data)
                dtype = type(data[0])
                if not all(isinstance(a, dtype) for a in data):
                    raise MetadataError("Data must have uniform type")
                else:
                    data = {'values': data}

            self._data = data
            self._fieldtypes = [type(data[k][0]) for k in data]
        return

    def __repr__(self):
        return "D[" + ", ".join(str(k) for k in self._data.keys()) + "]"

    def __add__(self, other):
        if not isinstance(other, type(self)):
            raise MetadataError("self and other must both be instances of "
                                "{0}".format(type(self)))
        if set(self.keys()) != set(other.keys()):
            raise MetadataError("self and other do not have identical field "
                                "names")
        res = Metadata({})
        res._len = self._len + other._len
        for k in self:
            res[k] = self[k] + other[k]
        return res

    def __len__(self):
        return self._len

    def __contains__(self, other):
        return (other in self._data)

    def __iter__(self):
        return self._data.__iter__()

    def __setitem__(self, key, value):
        if isinstance(key, (IntegerType, slice)):
            nkeys = len(self._data.keys())
            if (nkeys > 1) and (nkeys != len(list(value))):
                raise IndexError("Setting Metadata requires a value to be "
                                 "provided for every Metadata field")

            if nkeys > 1:
                for seq in self._data.values():
                    seq[key] = value[key]
            else:
                for k in self._data.keys():
                    self._data[k][key] = value

        else:
            if len(list(value)) != len(self):
                raise IndexError("Setting a Metadata field requires that "
                                 "length be preserved")
            self._data[key] = list(value)
        return

    def __getitem__(self, key):
        # If there is one field type, a number index should return a scalar
        # If there are multiple field types, a number index should return a dict
        # Although this is a bit irregular, it probably adheres better to the
        # principle of least surprise
        if isinstance(key, (IntegerType, slice)):
            return dict((k, self._data[k][key]) for k in self._data.keys())
        else:
            return self.getfield(key)

    def __delitem__(self, idx):
        for k in self._data:
            del self._data[k][idx]
        self._len -= 1

    def sub(self, idxs):
        """ Return a Metadata instance with values from idxs. """
        newdata = dict()
        try:
            items = self._data.iteritems()
        except AttributeError:
            items = self._data.items()

        for key, val in items:
            newdata[key] = [self._data[key][i] for i in idxs]
        return Metadata(newdata, copy=False)

    def getfield(self, name):
        """ Return all values from field *name*. """
        if name not in self._data:
            raise IndexError("Field {0} not in Metadata".format(name))
        return self._data[name]

    def update(self, other):
        """ Extend collection in-place. """
        if isinstance(other, type(self)):
            for i, k in enumerate(self._data):
                if type(self._fieldtypes[i]) == type(other._fieldtypes[i]):
                    self._data[k] += other._data[k]
                    self._len += other._len
                else:
                    raise MetadataError("Cannot combine metadata instances "
                                         "with different type hierarchies")
        return self

    def keys(self):
        """ Return internal dictionary keys. """
        return self._data.keys()

    def values(self):
        """ Return internal dictionary values. """
        return self._data.values()


class MetadataError(Exception):
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message





"""
Submodule for generating streamlines from rasters.
"""

from math import ceil, sqrt

def interpolate1(x, y, a, b, c, d):
    """ Return a value *v(x,y)* in the regular structured stencil

            a --- b
            |  v  |
            c --- d

    using linear interpolation. The coordinates (x, y) must be normalized by
    the horizontal and vertical grid spacings, respectively.
    """
    left = (c-a)*y + a
    right = (d-b)*y + b
    return (right - left) * x + left


def streamline2d(U, V, x0, y0, ds=0.5, max_nodes=5000, res=(1.0, 1.0),
                 tol=None, momentum=False):
    """ Integrate velocity field (*U*, *V*) using a 4th-order Runge-Kutta
    scheme, starting from *x0*, *y0*. A pair of lists with coordinates
    in X and Y is returned.

    *ds* is the step size

    *max_nodes* is the maximum number of steps to take. Iteration will be
    terminated automatically if the streamline reaches the edge of the vector
    field.

    *res* is the resolution of *U*, *V* in the same units as *x0*, *y0*.

    *tol* is the tolerance threshold, below which a streamline is considered
    stationary and aborted.

    *momentum* allows a streamline to continue moving across a plateau. (not
    implemented)
    """

    m, n = U.shape

    # Initialize streamline vectors
    X = [x0]
    Y = [y0]

    # Adjust the input arrays to make them have resolution 1x1
    U = U / res[0]
    V = V / res[1]

    # Put x0, y0 into grid units
    x0 = x0 / res[0]
    y0 = y0 / res[1]

    # Check that x0, y0 are within the bounds of U, V
    if (x0 >= n-1) or (y0 >= m-1) or (x0 < 0) or (y0 < 0):
        raise IndexError("starting position is outside vector field")

    i = 0

    while i < max_nodes:

        try:

            # Use linear interpolation to find dx/ds and dy/ds at (x0, y0)
            dxds = interpolate1(x0 % 1.0, y0 % 1.0,
                        U[y0, x0], U[y0, ceil(x0)],
                        U[ceil(y0), x0], U[ceil(y0), ceil(x0)])
            dyds = interpolate1(x0 % 1.0, y0 % 1.0,
                        V[y0, x0], V[y0, ceil(x0)],
                        V[ceil(y0), x0], V[ceil(y0), ceil(x0)])

            k1x = ds * dxds
            k1y = ds * dyds

            dxds = interpolate1(x0 % 1.0 + 0.5*ds, y0 % 1.0 + 0.5*k1x,
                        U[y0, x0], U[y0, ceil(x0)],
                        U[ceil(y0), x0], U[ceil(y0), ceil(x0)])
            dyds = interpolate1(x0 % 1.0 + 0.5*ds, y0 % 1.0 + 0.5*k1y,
                        V[y0, x0], V[y0, ceil(x0)],
                        V[ceil(y0), x0], V[ceil(y0), ceil(x0)])

            k2x = ds * dxds
            k2y = ds * dyds

            dxds = interpolate1(x0 % 1.0 + 0.5*ds, y0 % 1.0 + 0.5*k2x,
                        U[y0, x0], U[y0, ceil(x0)],
                        U[ceil(y0), x0], U[ceil(y0), ceil(x0)])
            dyds = interpolate1(x0 % 1.0 + 0.5*ds, y0 % 1.0 + 0.5*k2y,
                        V[y0, x0], V[y0, ceil(x0)],
                        V[ceil(y0), x0], V[ceil(y0), ceil(x0)])

            k3x = ds * dxds
            k3y = ds * dyds

            dxds = interpolate1(x0 % 1.0 + ds, y0 % 1.0 + k3x,
                        U[y0, x0], U[y0, ceil(x0)],
                        U[ceil(y0), x0], U[ceil(y0), ceil(x0)])
            dyds = interpolate1(x0 % 1.0 + ds, y0 % 1.0 + k3y,
                        V[y0, x0], V[y0, ceil(x0)],
                        V[ceil(y0), x0], V[ceil(y0), ceil(x0)])

            k4x = ds * dxds
            k4y = ds * dyds

            x0 = x0 + k1x/6.0 + k2x/3.0 + k3x/3.0 + k4x/6.0
            y0 = y0 + k1y/6.0 + k2y/3.0 + k3y/3.0 + k4y/6.0

        except IndexError:      # x0 or y0 was an invalid (e.g. nan)
            break

        # Check that x0, y0 are within the bounds of U, V
        if (x0 >= n-1) or (y0 >= m-1) or (x0 < 0) or (y0 < 0):
            break

        X.append(x0*res[0])
        Y.append(y0*res[1])

        # Check that rate of change is greater than the tolerance limit
        if tol is not None:
            if sqrt( k4x**2 + k4y**2 ) < tol:
                break

        i += 1

    return X, Y





""" Metadata tables for vector data """

from collections import Sequence

class Metadata(Sequence):

    def __init__(self, data, fields=None, checktypes=False):
        """ Create a collection of metadata from *data*.

        *data* may be:
        
            - a list with uniform type
            - a dictionary with equally-sized fields of uniform type.
            - a scalar
            - another Metadata instance

        The *kwarg* fields is used for fast Metadata object creation, and
        requires *data* to be provided as a list of equal-sized tuples.

        If *checktypes* is True (default False), the values in *data* will be
        tested to ensure they are of constant type. In the usual case, this
        type-checking is foregone for speed.
        """
        if fields is None:
            if hasattr(data, "_fields"):        # Data is Metadata-like
                self._fields = data._fields
                self._data = data._data
            elif hasattr(data, "keys"):         # Data is dict-like
                self._fields = tuple(data.keys())
                fst = data[self._fields[0]]
                if hasattr(fst, "__iter__") and not isinstance(fst, str):   # Vector entries
                    zipvalues = zip(*[data[f] for f in self._fields])
                    self._data = [tuple(item) for item in zipvalues]
                else:                           # Scalar entries
                    self._data = [(data[f],) for f in self._fields]
            else:                               # Data is list or scalar
                self._fields = ("value",)
                if hasattr(data, "__iter__") and not isinstance(data, str):
                    self._data = [(d,) for d in data]
                else:
                    self._data = [(data,)]
        else:
            if hasattr(data[0], "__len__") and not isinstance(data[0], str):
                if len(data[0]) != len(fields):
                    raise ValueError("Length of data entries and fields don't match")
            self._fields = tuple(fields)
            self._data = data

        if checktypes:
            self._validatetypes()
        return

    def _validatetypes(self):
        for i,t in enumerate(self.types):
            if not all(isinstance(t, d[i]) for d in self._data):
                raise TypeError("data contains item ({0}) not of type {1}".format(self._fields[i], t))
        return

    def __repr__(self):
        return "D[" + ", ".join(self._fields) + "]"

    def __eq__(self, other):
        return (self._data == other._data) and (self._fields == other._fields)

    def __neq__(self, other):
        return self != other

    def __getitem__(self, i):
        return self._data[i]

    def __setitem__(self, i, val):
        self._data[i] = val
        return

    def __delitem__(self, i):
        del self._data[i]

    def __len__(self):
        return len(self._data)

    def __contains__(self, field):
        return field in self._fields

    @property
    def data(self):
        return self._data

    @property
    def fields(self):
        return self._fields

    @property
    def types(self):
        return tuple(type(a) for a in self._data[0])

    def get(self, i):
        """ Return a dictionary for a single entry """
        d = self._data[i]
        if isinstance(i, slice):
            r = {}
            for j in range(len(self._fields)):
                r[self._fields[j]] = [d_[j] for d_ in d]
            return r
        else:
            return dict((self._fields[j], d[j]) for j in range(len(self._fields)))

    def getfield(self, field):
        """ Return list of data corresponding to *field* """
        if field in self._fields:
            i = self._fields.index(field)
            return [d[i] for d in self._data]
        else:
            raise KeyError("'{0}' not a field field".format(field))

    def extend(self, other):
        """ Extend Metadata from another Metadata instance. If *other* has
        field f in *self*, it is copied. Otherwise, the None is appended. """
        # TODO: appended None value should be approariate to the type of field f.
        idxs_other = []
        for f in self._fields:
            if f in other._fields:
                idxs_other.append(other._fields.index(f))
            else:
                idxs_other.append(None)

        for i in range(len(other)):
            self._data.append(tuple([None if idxs_other[j] is None
                                          else other._data[i][idxs_other[j]]
                                          for j in range(len(self._fields))]))




""" IO interface to GeoTiffs using GDAL. """

import sys
from osgeo import gdal, osr
import osgeo.gdalconst as gc
import numpy as np

gdal.UseExceptions()

def SRS_from_WKT(s):
    """ Return Proj.4 string, semimajor axis, and flattening """
    sr = osr.SpatialReference()
    sr.ImportFromWkt(s)
    return sr

def numpy_dtype(dt_int):
    """ Return a numpy dtype that matches the band data type. """
    name = gdal.GetDataTypeName(dt_int)
    if name == "UInt16":
        return np.uint16
    elif name == "Int16":
        return np.int16
    elif name == "UInt32":
        return np.uint32
    elif name == "Int32":
        return np.int32
    elif name == "Float32":
        return np.float32
    elif name == "Float64":
        return np.float64
    elif name in ("CInt16", "CInt32", "CFloat32", "CFloat64"):
        return np.complex64
    else:
        raise TypeError("GDAL data type {0} unknown to karta".format(dt_int))

def gdal_type(dtype):
    """ Return a GDAL type that matches numpy dtype """
    if dtype == np.uint16:
        return gdal.GDT_UInt16
    elif dtype == np.int16:
        return gdal.GDT_Int16
    elif dtype == np.int32:
        return gdal.GDT_Int32
    elif dtype == np.int32:
        return gdal.GDT_Int32
    elif dtype == np.float32:
        return gdal.GDT_Float32
    elif dtype == np.float64:
        return gdal.GDT_Float64
    elif dtype == np.complex64:
        return gdal.GDT_CFloat64
    else:
        raise TypeError("GDAL equivalent to type {0} unknown".format(dtype))

def read(fnm, band):
    """ Read a GeoTiff file and return a numpy array and a dictionary of header
    information. """
    hdr = dict()
    dataset = gdal.Open(fnm, gc.GA_ReadOnly)

    try:
        hdr["nx"] = dataset.RasterXSize
        hdr["ny"] = dataset.RasterYSize

        transform = dataset.GetGeoTransform()
        if transform is not None:
            hdr["dx"] = transform[1]
            hdr["dy"] = transform[5]
            hdr["xulcorner"] = transform[0]
            hdr["yulcorner"] = transform[3]
            hdr["sx"] = transform[2]
            hdr["sy"] = transform[4]
        else:
            raise AttributeError("No GeoTransform in geotiff file")

        sr = SRS_from_WKT(dataset.GetProjectionRef())
        hdr["srs"] = {"proj4": sr.ExportToProj4(),
                      "semimajor": sr.GetSemiMajor(),
                      "flattening": sr.GetInvFlattening()}

        max_dtype = 0
        rasterband = dataset.GetRasterBand(band)
        nx = rasterband.XSize
        ny = rasterband.YSize
        if rasterband.DataType > max_dtype:
            max_dtype = rasterband.DataType

        arr = np.empty((hdr["ny"], hdr["nx"]), dtype=numpy_dtype(max_dtype))
        dtype = numpy_dtype(rasterband.DataType)
        arr[:,:] = rasterband.ReadAsArray(buf_obj=np.empty([ny, nx], dtype=dtype))

    finally:
        dataset = None
    return arr, hdr

def srs_from_crs(crs):
    try:
        srs = osr.SpatialReference()
        srs.ImportFromProj4(crs.project.srs)
    except AttributeError:
        raise TypeError("Right now, GDAL projection can only be set from CRS instances backed by proj.4")
    return srs

def write(fnm, grid):
    """ Write a grid-like object to *fnm* """
    driver = gdal.GetDriverByName("GTiff")
    ny, nx = grid.size
    dataset = driver.Create(fnm, nx, ny, 1, gdal_type(grid.values.dtype))
    t = grid.transform
    dataset.SetGeoTransform([t[0] - 0.5*(t[2] + t[4]), t[2], -t[4],
                             t[1] + (ny - 0.5)*t[3] + 0.5*t[5], t[5], -t[3]])
    try:
        srs = srs_from_crs(grid.crs)
    except TypeError as e:
        sys.stderr.write("Writing GeoTiff failed:\n\t{0}\n".format(e))
        return
    dataset.SetProjection(srs.ExportToWkt())
    dataset.GetRasterBand(1).WriteArray(grid.values[::-1])
    dataset = None
    return




""" This module includes the `GeoJSON` class and convenience functions for
dealing with GeoJSON data. The `GeoJSON` class uses the builtin json module and
enforces GeoJSON standards. """

import copy
import json
from collections import namedtuple
from numbers import Number
from math import isnan
from functools import reduce
from ..crs import CRS

Point = namedtuple('Point', ['coordinates', 'crs'])
MultiPoint = namedtuple('MultiPoint', ['coordinates', 'crs'])
LineString = namedtuple('LineString', ['coordinates', 'crs'])
MultiLineString = namedtuple('MultiLineString', ['coordinates', 'crs'])
Polygon = namedtuple('Polygon', ['coordinates', 'crs'])
MultiPolygon = namedtuple('MultiPolygon', ['coordinates', 'crs'])
GeometryCollection = namedtuple('GeometryCollection', ['geometries'])
Feature = namedtuple('Feature', ['geometry', 'properties', 'id', 'crs'])
FeatureCollection = namedtuple('FeatureCollection', ['features', 'crs'])

class GeoJSONNamedCRS(CRS):
    def __init__(self, name):
        self.name = name
        self.jsonname = name

    def __eq__(self, other):
        return self.jsonname == other.jsonname

class GeoJSONLinkedCRS(CRS):
    def __init__(self, href, linktype):
        self.name = href
        self.jsonhref = href
        self.jsontype = linktype

    def __eq__(self, other):
        return self.jsonhref == other.jsonhref

DEFAULTCRS = GeoJSONNamedCRS("urn:ogc:def:crs:epsg::4326")

class ExtendedJSONEncoder(json.JSONEncoder):
    """ Numpy-specific numbers prior to 1.9 don't inherit from Python numeric
    ABCs. This class is a hack to coerce numpy values into Python types for
    JSON serialization. """

    def default(self, o):
        try:
            if o.dtype in ("int8", "int16", "int32", "int64"):
                return int(o)
            elif o.dtype in ("float16", "float32", "float64", "float128"):
                return float(o)
            elif o.dtype in ("complex64", "complex128", "complex256"):
                return complex(o)
            else:
                raise TypeError("not a recognized type")
        except (AttributeError, TypeError):
            return json.JSONEncoder.default(self, o)


class GeoJSONWriter(object):
    """ Class for converting geometry objects to GeoJSON strings. Multipoint-based
    opbjects are written as 'Features'.

    Notes:
    ------
    Does not handle 'FeatureCollection' types (see printFeatureCollection()
    function).
    """
    def __init__(self, gpobj, **kwargs):
        type_equiv = {'Point'       : 'Point',
                      'Multipoint'  : 'MultiPoint',
                      'Line'        : 'LineString',
                      'Polygon'     : 'Polygon'}

        if gpobj._geotype in type_equiv:
            self.typestr = type_equiv[gpobj._geotype]
        else:
            raise TypeError('Input object not a recognized geometry')
        crs = gpobj._crs
        bbox = kwargs.get('bbox', None)
        urn = kwargs.get('urn', None)

        self.gpobj = gpobj
        self.supobj = {}
        if crs is not None:
            self.add_crs(crs, urn=urn)
        if self.typestr != 'Point':
            self.supobj['type'] = 'Feature'
            self.add_bbox(bbox)
            self.add_geometry()
        else:
            self.supobj['type'] = 'Point'
            self.add_coordinates()
        return

    def add_crs(self, crs, **kw):
        """ Tag the JSON object with coordinate reference system metadata using
        a CRS instance, which is expected the have the interface of crs.CRS.

        In the case of a linked CRS, the link address and type can be specified
        using the `href` and `linktype` keyword arguments. If not supplied,
        then it is assumed that crs.id["urn"] contains the *href* and the
        *type* is "name".

        For more details, see the GeoJSON specification at:
        http://geojson.org/geojson-spec.html#coordinate-reference-system-objects
        """
        if hasattr(crs, "jsonhref") and hasattr(crs, "jsontype"):
            self.supobj['crs'] = {'type': 'link',
                                  'properties': {'href': crs.jsonname,
                                                 'type': crs.jsontype}}
        elif hasattr(crs, "jsonname"):
            self.supobj['crs'] = {'type': 'name',
                                  'properties': {'name': crs.jsonname}}
        else:
            urn = kw.get("urn", geturn(crs))
            self.supobj['crs'] = {'type': 'name',
                                  'properties': {'name': urn}}
        return

    def add_bbox(self, bbox=None):
        """ Tag the JSON object with a bounding box. If *bbox* is None
        (default), the bounds will be calculated from the data. Otherwise,
        *bbox* should be in the form (xmin, xmax, ....) for all dimensions.
        """
        if bbox is None:
            if hasattr(self.gpobj, 'get_extents'):
                xmn, xmx, ymn, ymx = self.gpobj.get_extents()
                bbox = ((xmn, xmx), (ymn, ymx))
            else:
                raise AttributeError("Type {0} has no 'get_extents' "
                                     "method".format(type(self.gpobj)))
        self.supobj['bbox'] = bbox
        return

    def add_geometry(self):
        """ Add geometry subobject. This should be called for writing
        Multipoint-derived classes. """
        self.supobj['geometry'] = {'type' : self.typestr}
        self.add_coordinates(self.supobj['geometry'])
        self.add_id()
        self.add_properties()
        return

    def add_coordinates(self, target=None):
        """ Add coordinates. Behaviour depends on the type of the geometry
        object. Target must be None (default) or a dictionary member of the
        JSON object. """
        if target is None:
            target = self.supobj

        if self.gpobj._geotype == 'Polygon':
            target['coordinates'] = [list_rec(self.gpobj.get_vertices())]
            if hasattr(self.gpobj, "subs"):
                for poly in self.gpobj.subs:
                    target['coordinates'].append(list_rec(poly.get_vertices()))
        elif hasattr(self.gpobj, 'get_vertices'):
            target['coordinates'] = list_rec(self.gpobj.get_vertices())
        elif hasattr(self.gpobj, 'get_vertex'):
            target['coordinates'] = list_rec(self.gpobj.get_vertex())
        else:
            raise AttributeError('Geometry object has no vertex method')
        return

    def add_properties(self, target=None):
        """ Add the data from the geometry object's data dictionary. """
        if target is None:
            target = self.supobj
        target['properties'] = {}
        if self.gpobj.data is not None:
            if hasattr(self.gpobj.data, 'fields'):
                data = self.gpobj.data
                data = dict((f, data.getfield(f)) for f in data.fields)
            else:
                data = {'point_data': self.gpobj.data}
            for key in data:
                nan2none = lambda a: None if isinstance(a, Number) and isnan(a) else a
                target['properties'][key] = list(map(nan2none, data[key]))
        return

    def add_id(self, target=None):
        """ Add an index field. """
        if target is None:
            target = self.supobj
        target['id'] = list(range(len(self.gpobj.get_vertices())))
        return

    def print_json(self):
        """ Print GeoJSON representation. """
        return json.dumps(self.supobj, indent=2, cls=ExtendedJSONEncoder)

    def write_json(self, fout):
        """ Dump internal dict-object to JSON using the builtin `json` module.
        """
        json.dump(self.supobj, fout, indent=2, cls=ExtendedJSONEncoder)
        return

class GeoJSONReader(object):

    def __init__(self, finput):
        """ Create a reader-object for a GeoJSON-containing file or StreamIO
        object. Use as::

            with open(`fnm`, 'r') as f:
                reader = GeoJSONReader(f)
        """
        if not hasattr(finput, 'read'):
            with open(finput) as f:
                self.jsondict = json.load(f)
        else:
            self.jsondict = json.load(finput)
        return

    @staticmethod
    def parsePoint(d, defaultcrs):
        crs = d.get("crs", defaultcrs)
        return Point(d["coordinates"], crs)

    @staticmethod
    def parseMultiPoint(d, defaultcrs):
        crs = d.get("crs", defaultcrs)
        return MultiPoint(d["coordinates"], crs)

    @staticmethod
    def parseLineString(d, defaultcrs):
        crs = d.get("crs", defaultcrs)
        return LineString(d["coordinates"], crs)

    @staticmethod
    def parseMultiLineString(d, defaultcrs):
        crs = d.get("crs", defaultcrs)
        return MultiLineString(d["coordinates"], crs)

    @staticmethod
    def parsePolygon(d, defaultcrs):
        crs = d.get("crs", defaultcrs)
        return Polygon(d["coordinates"], crs)

    @staticmethod
    def parseMultiPolygon(d, defaultcrs):
        crs = d.get("crs", defaultcrs)
        return MultiPolygon(d["coordinates"], crs)

    def parseGeometry(self, o, defaultcrs):
        crs = o.get("crs", defaultcrs)
        t = o["type"]
        if t == "Point":
            return self.parsePoint(o, crs)
        elif t == "MultiPoint":
            return self.parseMultiPoint(o, crs)
        elif t == "LineString":
            return self.parseLineString(o, crs)
        elif t == "MultiLineString":
            return self.parseMultiLineString(o, crs)
        elif t == "Polygon":
            return self.parsePolygon(o, crs)
        elif t == "MultiPolygon":
            return self.parseMultiPolygon(o, crs)
        else:
            raise TypeError("Unrecognized type {0}".format(t))

    def parseGeometryCollection(self, o, defaultcrs):
        crs = o.get("crs", defaultcrs)
        geoms = [self.parseGeometry(g) for g in o["geometries"]]
        return GeometryCollection(geoms, crs)

    def parseFeature(self, o, defaultcrs):
        crs = o.get("crs", defaultcrs)
        geom = self.parseGeometry(o["geometry"], crs)
        prop = self.parseProperties(o["properties"], len(geom.coordinates))
        fid = o.get("id", None)
        return Feature(geom, prop, fid, crs)

    def parseFeatureCollection(self, o, defaultcrs):
        crs = o.get("crs", defaultcrs)
        features = [self.parseFeature(f, crs) for f in o["features"]]
        return FeatureCollection(features, crs)

    @staticmethod
    def parseProperties(prop, geomlen):
        d = {"scalar":{},
             "vector":{}}
        for key, value in prop.items():
            if geomlen > 1 and hasattr(value, "__iter__") and len(value) == geomlen:
                d["vector"][key] = value
            else:
                d["scalar"][key] = value
        return d

    def items(self, o=None):
        if o is None:
            o = self.jsondict
        crs = o.get("crs", DEFAULTCRS)
        items = []
        if o["type"] == "GeometryCollection":
            items.append(self.parseGeometryCollection(o, crs))
        elif o["type"] == "FeatureCollection":
            items.append(self.parseFeatureCollection(o, crs))
        elif o["type"] == "Feature":
            items.append(self.parseFeature(o, crs))
        else:
            items.append(self.parseGeometry(o, crs))
        return items

def geturn(crs):
    """ Return a urn string for *crs* """
    print("Warning: CRS URN lookup not implemented")
    print(crs)
    print("returning CRS name instead")
    return crs.name


def list_rec(A):
    """ Recursively convert nested iterables to nested lists """
    if hasattr(A, '__iter__'):
        return [list_rec(el) for el in A]
    else:
        return A

def printGeometryCollection(gpobj_list, **kwargs):
    """ Given an iterable that returns geometry objects, construct a GeoJSON
    FeatureCollection string.
    """
    geometrylist = []
    for gp in gpobj_list:
        writer = GeoJSONWriter(gp)
        geometrylist.append(copy.copy(writer.supobj))
    geometry_coll = {'type' : 'GeometryCollection',
                    'geometries' : geometrylist}
    kwargs.setdefault("indent", 2)
    return json.dumps(geometry_coll, **kwargs)

def geojson2csv(fin, fout):
    """ Convert a JSON file to a CSV. Only properties (attributes) common to
    all features are retained. The arguments `fin` and `fout` may be either
    filenames or file-like objects.
    """

    if hasattr(fin, 'read'):
        jsdat = json.load(fin)
    else:
        with open(fin, 'r') as f:
            jsdat = json.load(f)

    # Load coordinates and properties
    X = []
    Y = []
    PROP = []
    for feat in jsdat['features']:
        x, y = feat['geometry']['coordinates']
        X.append(x)
        Y.append(y)
        PROP.append(feat['properties'])

    # Find subset of properties that exist in all features
    if len(PROP) > 1:
        s = set(PROP[0].keys())
        common_props = s.intersection(*[set(p.keys()) for p in PROP[1:]])
    else:
        common_props = PROP[0].keys()


    # Output
    outstr = 'x,y' + \
             reduce(lambda a,b: a+','+b, common_props) + \
             '\n'   # Construct header

    outstr += reduce(lambda a, b: a+'\n'+b,                         # Combine lines
                map(lambda x, y, p: str(x) + ',' + str(y) + ',' +   # Construct lines
                    reduce(lambda a, b: str(a) + ',' + str(b),      # Construct properties
                        [p[cp] for cp in common_props]),            # Filter spurious props
                    X, Y, PROP))                                    # Args for line lambda

    try:
        if hasattr(fout, 'read'):
            f = fout
        else:
            f = open(fout, 'w')
        f.write(outstr)

    finally:
        if not hasattr(fout, 'read'):
            f.close()

    return





""" Functions for reading the USGS DEM format. """

import math
import itertools
import re

def coerce_float(s):
    return float(s.replace("D","e",1)
                  .replace("E","e",1)
                  .replace("F","e",1)
                  .replace("G","e",1))

FMT_KEY = {"A":str,
           "I":int,
           "D":coerce_float,
           "E":coerce_float,
           "F":coerce_float,
           "G":coerce_float}


BLOCKA = [("fnm",0,40,"A40"),
          ("free",40,109,"A69"),
          ("se_corner",109,135,"2(I4,I2,F7.4)"),
          ("proc_code",135,136,"A1",),
          ("sectional_indicator",137,140,"A3"),
          ("origin_code",140,144,"A4"),
          ("dem_level_code",144,150,"I6"),
          ("elev_pattern",150,156,"I6"),
          ("crs",156,162,"I6"),
          ("crs_zone",162,168,"I6"),
          ("proj",168,528,"15D24.15"),
          ("crs_unit",528,534,"I6"),
          ("z_unit",534,540,"I6"),
          ("nsides",540,546,"I6"),
          ("quad_coords",546,738,"4(2D24.15)"),
          ("minmax",738,786,"2D24.15"),
          ("angle_ccw",786,810,"D24.15"),
          ("z_acc",810,816,"I6"),
          ("resolution",816,852,"3E12.6"),
          ("size",852,864,"2I6"),       # Old format end
          ("max_contour_int",864,869,"I5"),
          ("max_contour_units",869,870,"I1"),
          ("min_contour_int",870,875,"I5"),
          ("min_contour_units",875,876,"I1"),
          ("src_year",876,880,"I4"),
          ("rev_year",880,884,"I4"),
          ("inspection_flag",884,885,"A1"),
          ("validation_flag",885,886,"A1"),
          ("void_flag",886,888,"I1"),
          ("z_datum",888,890,"I2"),
          ("xy_datum",890,892,"I2"),
          ("edition",892,896,"I4"),
          ("void_perc",896,900,"I4"),
          ("edge_flag",900,908,"4I2"),
          ("z_shift",908,915,"F7.2")]

BLOCKB = [("rc_id",0,12,"2I6"),
          ("mn",12,24,"2I6"),
          ("xy",24,72,"2D24.15"),
          ("zdatum",72,96,"D24.15"),
          ("zminmax",96,144,"2D24.15")]

BLOCKC = [("filestatcode",0,6,"I6"),
          ("file_rmse",6,24,"3I6"),
          ("file_nsmp",24,30,"I6"),
          ("demstatcode",30,36,"I6"),
          ("dem_rmse",36,54,"3I6"),
          ("dem_nsmp",54,60,"I6")]

def nreps(fmt):
    """ Return the number of characters in a single record of
    a potentially multiple-record string. """
    try:
        reps = int(fmt[0])
        return reps * nreps(fmt[1:].lstrip("(").rstrip(")"))

    except ValueError:
        return 1
        #return fmt.count(",") + 1

def nreps_re(fmt):
    M = re.match(r"\A\d+", fmt)
    if M:
        return int(fmt[:M.end()]) * nreps_re(fmt[M.end():].lstrip("(").rstrip(")"))
    else:
        return 1

def reclen(fmt):
    """ Return the number of characters in a single record of
    a potentially multiple-record string. """
    try:
        reps = int(fmt[0])
        return reclen(fmt[1:].lstrip("(").rstrip(")"))

    except ValueError:
        if "," in fmt:
            return list(map(lambda a: reclen(a)[0], fmt.split(",")))

        if fmt[0] in ("G", "F", "E", "D"):
            nch = int(fmt[1:].split(".")[0])
        elif fmt[0] in ("A", "I"):
            nch = int(fmt[1:])
        else:
            raise ValueError("Unrecognized format code {0}"
                             .format(fmt))
        return [nch]

def dtype(fmt):
    """ Return a constructor for the datatype encoded in *fmt*. """
    try:
        reps = int(fmt[0])
        return dtype(fmt[1:].lstrip("(").rstrip(")"))

    except ValueError:
        if "," in fmt:
            return list(map(lambda a: dtype(a)[0], fmt.split(",")))
        else:
            return [FMT_KEY[fmt[0]]]

def cumsum(u):
    if len(u) == 1:
        return u
    elif len(u) == 2:
        return [u[0], u[0]+u[1]]
    else:
        return [u[0]] + cumsum([u[0]+u[1]] + u[2:])

def cumsum_loop(u):
    v = [u[0]]
    for i in u[1:]:
        v.append(v[-1] + i)
    return v

def parse(fmt, s):
    """ Based on an ASCII format string, parse the information in *s*. """
    nch = reclen(fmt)
    reps = nreps_re(fmt)
    t = dtype(fmt)
    pos = [0] + cumsum(nch*reps)
    vals = [t_(s[a:b]) for a,b,t_ in zip(pos[:-1], pos[1:],
                                             itertools.cycle(t))]
    return vals if len(vals) > 1 else vals[0]

def read_block(blk, KEY):
    """ Read a block from a USGS .dem file given *KEY*, which is a list of
    tuples in (name::str, pos::int, end::int, format::str) form. """
    if len(blk) > 1024:
        blk = blk[:1024]
    blkdict = {}
    for field,pos0,pos1,fmt in KEY[:20]:
        blkdict[field] = parse(fmt, blk[pos0:pos1])
    return blkdict

def demread(fnm):
    """ Read a USGS (or CDED) .dem file and return a dictionary for block A, a list of raw data values, and a dictionary for block B. """
    with open(fnm, "r") as f:
        data = f.read()

    blocka = read_block(data[:1024], BLOCKA)

    dem = []
    profnz = parse("2I6", data[1036:1048])[0]
    profnch = profnz * 6
    for profnum in range(blocka["size"][1]):
        i = (int(math.ceil((profnch-146) / 1024.)) + 1) * profnum + 1
        bhdr = read_block(data[i*1024:(i + 1) * 1024], BLOCKB)
        #profdem = read_block(data[1168:2048], [("data",0,876,"146I6")])["data"]
        profdem = read_block(data[1024*i:1024*(i+1)],
                             [("data",144,1020,"146I6")])["data"]
        blknz = [170 for _ in range((profnz-146)//170)] + [(profnz-146) % 170]
        for j, nz in enumerate(blknz):
            fmt = str(nz) + "I6"
            profdem.extend(read_block(data[1024*(j+i+1) : 1024*(j+i+2)],
                                      [("data",0,1020,fmt)])["data"])
        dem.extend(profdem)

    blockc = read_block(data[-1024:], BLOCKC)
    return blocka, dem, blockc




""" Low-level functions for reading ESRI ASCII grids """

import numpy as np

def aairead(fnm):
    """ Read an existing ASCII grid file and return a Numpy array and a
    dictionary of header information. """
    try:
        h = []
        with open(fnm, 'r') as f:
            # Count the number of header entries
            cnt = 0
            while True:
                l = f.readline()
                if l.split(None, 1)[0].lower() in ['nrows', 'ncols',
                        'yllcenter', 'xllcenter', 'yllcorner', 'xllcorner',
                        'cellsize', 'nodata_value']:
                    cnt += 1
                else:
                    break

            # Read the header, then the array data
            f.seek(0)
            for _ in range(cnt):
                h.append(f.readline())
            data = f.readlines()

    except IOError:
        raise AAIIOError('error while trying to open {0}'.format(fnm))

    # Store data internally
    hs = [rec.split(None, 1) for rec in h]
    hdr = dict([(rec[0].lower(), float(rec[1])) for rec in hs])
    hdr['ncols'] = int(hdr['ncols'])
    hdr['nrows'] = int(hdr['nrows'])
    if 'yllcenter' not in hdr.keys():
        hdr['yllcenter'] = None

    if 'xllcenter' not in hdr.keys():
        hdr['xllcenter'] = None

    if 'yllcorner' not in hdr.keys():
        hdr['yllcorner'] = None

    if 'xllcorner' not in hdr.keys():
        hdr['xllcorner'] = None

    if 'nodata_value' not in hdr.keys():
        hdr['nodata_value'] = -9999

    check_header(hdr)

    f = lambda l: [float(i) for i in l.split()]
    data_f = list(map(f, data))
    data_a = np.array(data_f)
    data_a[data_a==hdr['nodata_value']] = np.nan

    return data_a, hdr


def check_header(hdr):
    """ Make sure that all required header records are present, as well as both
    centered and corner spatial references. """
    for field in ('ncols', 'nrows', 'cellsize'):
        if field not in hdr:
            raise AAIError("{0} not defined".format(field.upper()))

    d = hdr['cellsize']

    if hdr.get('yllcenter') is None:
        if hdr.get('yllcorner') is None:
            raise AAIIOError('YLL reference not defined')
        else:
            hdr['yllcenter'] = hdr['yllcorner'] + d / 2.0
    else:
        hdr['yllcorner'] = hdr['yllcenter'] - d / 2.0

    if hdr.get('xllcenter') is None:
        if hdr.get('xllcorner') is None:
            raise AAIIOError('XLL reference not defined')
        else:
            hdr['xllcenter'] = hdr['xllcorner'] + d / 2.0
    else:
        hdr['xllcorner'] = hdr['xllcenter'] - d / 2.0

    return hdr


class AAIError(Exception):
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message


class AAIIOError(Exception):
    """ Exceptions related to ESRI ASCII grid driver. """
    def __init__(self, value, detail=None):
        self.value = value
        self.detail = detail
    def __str__(self):
        if self.detail is not None:
            s = self.value + ': ' + self.detail
        else:
            s = self.value
        return repr(s)



from math import ceil, sqrt
import numpy as np
cimport numpy as np
from cpython cimport bool

def diffrad(double a, double b):
    """ Return the difference in radians between two angles """
    cdef double pi
    pi = 3.141592653589793
    return (a + pi - b) % (2*pi) - pi

cdef float interpolate1(float x, float y, float a, float b, float c, float d):
    """ Return a value *v(x,y)* in the regular structured stencil
    
            a --- b
            |  v  |
            c --- d
    
    using linear interpolation. The coordinates (x, y) must be normalized by
    the horizontal and vertical grid spacings, respectively.
    """
    cdef float left, right, res
    
    left = (c-a)*y + a
    right = (d-b)*y + b
    res = (right - left) * x + left
    return res
    
def streamline2d(np.ndarray U, np.ndarray V, float x0, float y0,
                 float ds=0.5, int max_nodes=5000, tuple res=(1.0, 1.0),
                 float tol=0.0, int momentum=False):
    cdef int m, n, i
    cdef list X, Y
    cdef float dxds, dyds
    cdef float k1x, k1y, k2x, k2y, k3x, k3y, k4x, k4y
    cdef tuple result
    
    m = U.shape[0]
    n = U.shape[1]

    # Initialize streamline vectors
    X = [x0]
    Y = [y0]

    # Adjust the input arrays to make them have resolution 1x1
    U = U / res[0]
    V = V / res[1]

    # Put x0, y0 into grid units
    x0 = x0 / res[0]
    y0 = y0 / res[1]

    # Check that x0, y0 are within the bounds of U, V
    if (x0 >= n-1) or (y0 >= m-1) or (x0 < 0) or (y0 < 0):
        raise IndexError("starting position is outside vector field")

    i = 0

    while i < max_nodes:

        try:

            # Use linear interpolation to find dx/ds and dy/ds at (x0, y0)
            dxds = interpolate1(x0 % 1.0, y0 % 1.0,
                        U[y0, x0], U[y0, ceil(x0)],
                        U[ceil(y0), x0], U[ceil(y0), ceil(x0)])
            dyds = interpolate1(x0 % 1.0, y0 % 1.0,
                        V[y0, x0], V[y0, ceil(x0)],
                        V[ceil(y0), x0], V[ceil(y0), ceil(x0)])

            k1x = ds * dxds
            k1y = ds * dyds

            dxds = interpolate1(x0 % 1.0 + 0.5*ds, y0 % 1.0 + 0.5*k1x,
                        U[y0, x0], U[y0, ceil(x0)],
                        U[ceil(y0), x0], U[ceil(y0), ceil(x0)])
            dyds = interpolate1(x0 % 1.0 + 0.5*ds, y0 % 1.0 + 0.5*k1y,
                        V[y0, x0], V[y0, ceil(x0)],
                        V[ceil(y0), x0], V[ceil(y0), ceil(x0)])

            k2x = ds * dxds
            k2y = ds * dyds

            dxds = interpolate1(x0 % 1.0 + 0.5*ds, y0 % 1.0 + 0.5*k2x,
                        U[y0, x0], U[y0, ceil(x0)],
                        U[ceil(y0), x0], U[ceil(y0), ceil(x0)])
            dyds = interpolate1(x0 % 1.0 + 0.5*ds, y0 % 1.0 + 0.5*k2y,
                        V[y0, x0], V[y0, ceil(x0)],
                        V[ceil(y0), x0], V[ceil(y0), ceil(x0)])

            k3x = ds * dxds
            k3y = ds * dyds

            dxds = interpolate1(x0 % 1.0 + ds, y0 % 1.0 + k3x,
                        U[y0, x0], U[y0, ceil(x0)],
                        U[ceil(y0), x0], U[ceil(y0), ceil(x0)])
            dyds = interpolate1(x0 % 1.0 + ds, y0 % 1.0 + k3y,
                        V[y0, x0], V[y0, ceil(x0)],
                        V[ceil(y0), x0], V[ceil(y0), ceil(x0)])

            k4x = ds * dxds
            k4y = ds * dyds

            x0 = x0 + k1x/6.0 + k2x/3.0 + k3x/3.0 + k4x/6.0
            y0 = y0 + k1y/6.0 + k2y/3.0 + k3y/3.0 + k4y/6.0

        except IndexError:      # x0 or y0 was an invalid (e.g. nan)
            break

        # Check that x0, y0 are within the bounds of U, V
        if (x0 >= n-1) or (y0 >= m-1) or (x0 < 0) or (y0 < 0):
            break

        X.append(x0*res[0])
        Y.append(y0*res[1])

        # Check that rate of change is greater than the tolerance limit
        if tol > 0.0:
            if sqrt( k4x**2 + k4y**2 ) < tol:
                break

        i += 1

    result = (X, Y)
    return result
    
    



""" Creates a database of CRS IDs. The CRSRegister class tracks a dictionary of
CRS instances. When a new CRS is requested, it is returned from the dictionary
or created on the fly.

Customized reference systems may be added at runtime as follows:

::

    from karta.crs import crsreg

    crsreg.add_CRS("my_CRS", proj=[projection parameters],
                             geod=[geod parameters],
                             crstype=[one of 'projected', 'local', or 'geographical'],
                             id=[OGC URN (string)])

For example, to add a CRS for handling data in the European Terrestrial
Reference System of 1989, use:

::

    crsreg.add_CRS("ETRS89", proj={"proj": "lonlat", "towgs84": [0,0,0,0,0,0,0]},
                             geod={"ellps": "GRS80"},
                             crstype="geographical",
                             id="urn:ogc:def:crs:EPSG::4258")
"""

import pyproj

class CRS(object):

    """ The `CRS` class represents a specific coordinate reference systems.
    Each `CRS` instance has a `crstype`, which is one of:

        - "geographical"
        - "projected"
        - "local"

    and an `id` string, which is intended to be an Open Geospatial Consortium
    uniform resource name ('uri'). Also, a `CRS` may have a callable `proj`
    method that projects between the `CRS` and geographical longitude and
    latitude, and a callable `geod` attribute that computes geodetic distances.
    """
    def __init__(self, proj=None, geod=None, crstype="unkown", id=None):
        """'proj' : pyproj.Proj or None

        'geod' : pyproj.Geod or None

        'crstype' : string
        Must be one of 'geographical', 'projected', or 'local'

        'id' : string
        Open Geospatial Consortium uniform resource name
        """
        if crstype in ("geographical", "projected", "local", "unknown"):
            self.crstype = crstype
        else:
            raise TypeError("CRS argument 'crstype' must be 'geographical', "
                            "'projected', 'local', or 'unknown'")
        self.proj = proj
        self.geod = geod
        self.id = id
        return

    def __repr__(self):
        return "{0} CRS (urn: {1})".format(self.crstype, self.id)

    def __str__(self):
        return "{0} CRS (ID: {1}".format(self.crstype, self.id)

    def __eq__(self, other):
        if (self.crstype is not "unknown" and self.crstype == other.crstype and
            self.id == other.id and
            getattr(self.proj, "srs", 0) == getattr(other.proj, "srs", 0) and
            getattr(self.geod, "initstring", 0) == getattr(other.geod, "initstring", 0)):
                return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

class CRSRegister(object):
    """ The `CRSRegister` is a database of known and initialized coordinate
    systems. Individual `CRS` instances can be accessed as attributes of the
    `CRSRegister`. If they are already in use, a reference to the existing
    `CRS` is returned to avoid duplication of thousands of `CRS` instances.
    Otherwise, known references systems are initialized and returned.
    """
    
    def __init__(self):
        self.register = {}

    def __getattr__(self, name):
        return self.get_CRS(name)

    def add_CRS(self, name, **kwargs):
        """ Add a ``crs.CRS`` instance with `**kwargs` with the identifier
        `name`. """
        self.register[name] = CRS(**kwargs)

    def get_CRS(self, name):
        """ Get a reference to a ``crs.CRS`` with identifier `name`.
        """
        crs = self.register.get(name, None)
        if crs is None:
            if name in CRSDict:
                cfg = CRSDict[name]

                if cfg["proj"] is None:
                    p = None
                else:
                    p = pyproj.Proj(**cfg["proj"])

                if cfg["geod"] is None:
                    g = None
                else:
                    g = pyproj.Geod(**cfg["geod"])

                crs = CRS(proj=p, geod=g, crstype=cfg["crstype"], id=cfg["id"])
                self.register[name] = crs
            else:
                raise CRSLookupError("CRS '{0}' not defined".format(name))
        return crs

CRSDict = {"CARTESIAN" : {"proj": None,
                          "geod": None,
                          "crstype": "local",
                          "id": {"urn": "urn:ogc:def:crs:EPSG::5806"}},

           # Geographical
           "LONLAT_WGS84" : {"proj": {"proj": "lonlat"},
                             "geod": {"ellps":"WGS84"},
                             "crstype": "geographical",
                             "id": {"urn": "urn:ogc:def:crs:EPSG::4326"}},
           "LONLAT_NAD27" : {"proj": {"proj": "lonlat"},
                             "geod": {"ellps":"clrk66"},
                             "crstype": "geographical",
                             "id": {"urn": "urn:ogc:def:crs:EPSG::4267"}},
           "LONLAT_NAD83" : {"proj": {"proj": "lonlat"},
                             "geod": {"ellps":"GRS80"},
                             "crstype": "geographical",
                             "id": {"urn": "urn:ogc:def:crs:EPSG::4269"}},

           # Polar stereographic
           "UPSNORTH_WGS84" : {"proj": {"proj": "stere",
                                        "lat_0": 90,
                                        "lat_ts": 90,
                                        "lon_0": 0,
                                        "k": 0.994,
                                        "x_0": 2000000,
                                        "y_0": 2000000,
                                        "units": "m",
                                        "no_defs": True},
                               "geod": {"ellps": "WGS84"},
                               "crstype": "projected",
                               "id": {"urn": "urn:ogc:def:crs:EPSG::32661"}},
           "UPSSOUTH_WGS84" : {"proj": {"proj": "stere",
                                        "lat_0": -90,
                                        "lat_ts": -90,
                                        "lon_0": 0,
                                        "k": 0.994,
                                        "x_0": 2000000,
                                        "y_0": 2000000,
                                        "units": "m",
                                        "no_defs": True},
                               "geod": {"ellps": "WGS84"},
                               "crstype": "projected",
                   "id": {"urn": "urn:ogc:def:crs:EPSG::32761"}},
           "NSIDCNORTH_WGS84" : {"proj": {"proj": "stere",
                                          "lat_0": 90,
                                          "lat_ts": 70,
                                          "lon_0": -45,
                                          "k": 1,
                                          "x_0": 0,
                                          "y_0": 0,
                                          "units": "m",
                                          "no_defs": True},
                                 "geod": {"ellps": "WGS84"},
                                 "crstype": "projected",
                                 "id": {"urn": "urn:ogc:def:crs:EPSG::3413"}},
           "NSIDCSOUTH_WGS84" : {"proj": {"proj": "stere",
                                          "lat_0": -90,
                                          "lat_ts": -70,
                                          "lon_0": 0,
                                          "k": 1,
                                          "x_0": 0,
                                          "y_0": 0,
                                          "units": "m",
                                          "no_defs": True},
                                 "geod": {"ellps": "WGS84"},
                                 "crstype": "projected",
                                 "id": {"urn": "urn:ogc:def:crs:EPSG::3976"}},

           # Unknown
           "UNKNOWN" : {"proj": None,
                        "geod": None,
                        "crstype": "local",
                        "id": {}}
           }

# Aliases
CRSDict["LONLAT"] = CRSDict["LONLAT_WGS84"]
CRSDict["UPSNORTH"] = CRSDict["UPSNORTH_WGS84"]
CRSDict["UPSSOUTH"] = CRSDict["UPSSOUTH_WGS84"]
CRSDict["NSIDCNORTH"] = CRSDict["NSIDCNORTH_WGS84"]
CRSDict["NSIDCSOUTH"] = CRSDict["NSIDCSOUTH_WGS84"]

# This is the CRS register used globally within karta
crsreg = CRSRegister()

# UTM zones
#UTM1N = CRS("UTM Zone 1 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32601")
#UTM1S = CRS("UTM Zone 1 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32701")
#UTM2N = CRS("UTM Zone 2 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32602")
#UTM2S = CRS("UTM Zone 2 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32702")
#UTM3N = CRS("UTM Zone 3 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32603")
#UTM3S = CRS("UTM Zone 3 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32703")
#UTM4N = CRS("UTM Zone 4 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32604")
#UTM4S = CRS("UTM Zone 4 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32704")
#UTM5N = CRS("UTM Zone 5 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32605")
#UTM5S = CRS("UTM Zone 5 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32705")
#UTM6N = CRS("UTM Zone 6 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32606")
#UTM6S = CRS("UTM Zone 6 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32706")
#UTM7N = CRS("UTM Zone 7 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32607")
#UTM7S = CRS("UTM Zone 7 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32707")
#UTM8N = CRS("UTM Zone 8 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32608")
#UTM8S = CRS("UTM Zone 8 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32708")
#UTM9N = CRS("UTM Zone 9 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32609")
#UTM9S = CRS("UTM Zone 9 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32709")
#UTM10N = CRS("UTM Zone 10 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32610")
#UTM10S = CRS("UTM Zone 10 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32710")
#UTM11N = CRS("UTM Zone 11 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32611")
#UTM11S = CRS("UTM Zone 11 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32711")
#UTM12N = CRS("UTM Zone 12 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32612")
#UTM12S = CRS("UTM Zone 12 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32712")
#UTM13N = CRS("UTM Zone 13 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32613")
#UTM13S = CRS("UTM Zone 13 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32713")
#UTM14N = CRS("UTM Zone 14 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32614")
#UTM14S = CRS("UTM Zone 14 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32714")
#UTM15N = CRS("UTM Zone 15 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32615")
#UTM15S = CRS("UTM Zone 15 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32715")
#UTM16N = CRS("UTM Zone 16 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32616")
#UTM16S = CRS("UTM Zone 16 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32716")
#UTM17N = CRS("UTM Zone 17 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32617")
#UTM17S = CRS("UTM Zone 17 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32717")
#UTM18N = CRS("UTM Zone 18 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32618")
#UTM18S = CRS("UTM Zone 18 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32718")
#UTM19N = CRS("UTM Zone 19 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32619")
#UTM19S = CRS("UTM Zone 19 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32719")
#UTM20N = CRS("UTM Zone 20 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32620")
#UTM20S = CRS("UTM Zone 20 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32720")
#UTM21N = CRS("UTM Zone 21 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32621")
#UTM21S = CRS("UTM Zone 21 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32721")
#UTM22N = CRS("UTM Zone 22 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32622")
#UTM22S = CRS("UTM Zone 22 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32722")
#UTM23N = CRS("UTM Zone 23 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32623")
#UTM23S = CRS("UTM Zone 23 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32723")
#UTM24N = CRS("UTM Zone 24 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32624")
#UTM24S = CRS("UTM Zone 24 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32724")
#UTM25N = CRS("UTM Zone 25 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32625")
#UTM25S = CRS("UTM Zone 25 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32725")
#UTM26N = CRS("UTM Zone 26 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32626")
#UTM26S = CRS("UTM Zone 26 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32726")
#UTM27N = CRS("UTM Zone 27 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32627")
#UTM27S = CRS("UTM Zone 27 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32727")
#UTM28N = CRS("UTM Zone 28 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32628")
#UTM28S = CRS("UTM Zone 28 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32728")
#UTM29N = CRS("UTM Zone 29 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32629")
#UTM29S = CRS("UTM Zone 29 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32729")
#UTM30N = CRS("UTM Zone 30 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32630")
#UTM30S = CRS("UTM Zone 30 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32730")
#UTM31N = CRS("UTM Zone 31 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32631")
#UTM31S = CRS("UTM Zone 31 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32731")
#UTM32N = CRS("UTM Zone 32 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32632")
#UTM32S = CRS("UTM Zone 32 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32732")
#UTM33N = CRS("UTM Zone 33 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32633")
#UTM33S = CRS("UTM Zone 33 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32733")
#UTM34N = CRS("UTM Zone 34 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32634")
#UTM34S = CRS("UTM Zone 34 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32734")
#UTM35N = CRS("UTM Zone 35 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32635")
#UTM35S = CRS("UTM Zone 35 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32735")
#UTM36N = CRS("UTM Zone 36 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32636")
#UTM36S = CRS("UTM Zone 36 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32736")
#UTM37N = CRS("UTM Zone 37 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32637")
#UTM37S = CRS("UTM Zone 37 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32737")
#UTM38N = CRS("UTM Zone 38 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32638")
#UTM38S = CRS("UTM Zone 38 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32738")
#UTM39N = CRS("UTM Zone 39 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32639")
#UTM39S = CRS("UTM Zone 39 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32739")
#UTM40N = CRS("UTM Zone 40 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32640")
#UTM40S = CRS("UTM Zone 40 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32740")
#UTM41N = CRS("UTM Zone 41 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32641")
#UTM41S = CRS("UTM Zone 41 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32741")
#UTM42N = CRS("UTM Zone 42 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32642")
#UTM42S = CRS("UTM Zone 42 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32742")
#UTM43N = CRS("UTM Zone 43 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32643")
#UTM43S = CRS("UTM Zone 43 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32743")
#UTM44N = CRS("UTM Zone 44 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32644")
#UTM44S = CRS("UTM Zone 44 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32744")
#UTM45N = CRS("UTM Zone 45 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32645")
#UTM45S = CRS("UTM Zone 45 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32745")
#UTM46N = CRS("UTM Zone 46 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32646")
#UTM46S = CRS("UTM Zone 46 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32746")
#UTM47N = CRS("UTM Zone 47 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32647")
#UTM47S = CRS("UTM Zone 47 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32747")
#UTM48N = CRS("UTM Zone 48 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32648")
#UTM48S = CRS("UTM Zone 48 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32748")
#UTM49N = CRS("UTM Zone 49 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32649")
#UTM49S = CRS("UTM Zone 49 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32749")
#UTM50N = CRS("UTM Zone 50 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32650")
#UTM50S = CRS("UTM Zone 50 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32750")
#UTM51N = CRS("UTM Zone 51 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32651")
#UTM51S = CRS("UTM Zone 51 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32751")
#UTM52N = CRS("UTM Zone 52 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32652")
#UTM52S = CRS("UTM Zone 52 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32752")
#UTM53N = CRS("UTM Zone 53 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32653")
#UTM53S = CRS("UTM Zone 53 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32753")
#UTM54N = CRS("UTM Zone 54 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32654")
#UTM54S = CRS("UTM Zone 54 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32754")
#UTM55N = CRS("UTM Zone 55 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32655")
#UTM55S = CRS("UTM Zone 55 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32755")
#UTM56N = CRS("UTM Zone 56 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32656")
#UTM56S = CRS("UTM Zone 56 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32756")
#UTM57N = CRS("UTM Zone 57 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32657")
#UTM57S = CRS("UTM Zone 57 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32757")
#UTM58N = CRS("UTM Zone 58 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32658")
#UTM58S = CRS("UTM Zone 58 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32758")
#UTM59N = CRS("UTM Zone 59 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32659")
#UTM59S = CRS("UTM Zone 59 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32759")
#UTM60N = CRS("UTM Zone 60 North (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32660")
#UTM60S = CRS("UTM Zone 60 South (WGS 84)", "projected", "urn:ogc:def:crs:EPSG::32760")

class CRSError(Exception):
    """ Exception to raise for invalid geodetic operations. """
    def __init__(self, message=''):
        self.message = message

 
class CRSLookupError(Exception):
    def __init__(self, message=''):
        self.message = message




"""
Vector data classes and functions.
"""

from . import geometry
from . import xyfile
from . import geojson
from . import gpx
from . import vtk
from . import quadtree

from .geometry import Point, Multipoint, Line, Polygon
from .read import read_geojson, read_shapefile, from_shape
from .shp import write_shapefile
from .gpx import GPX
from .xyfile import load_xy, xyz2array_reg, array2xyz

__all__ = ["geometry", "xyfile", "geojson", "gpx", "vtk",
           "Point", "Multipoint", "Line", "Polygon",
           "read_geojson", "read_shapefile",
           "write_shapefile",
           "GPX",
           "load_xy", "xyz2array_reg", "array2xyz"]




""" Defines basic geodetic operations on a planes and spheres. """

#
# Functions are defined with an underscored version that works on scalars and a
# public version that dispatches between scalars and vectors. This approach is
# used rather than vectorization to make the C translation more straighforward.
#

import numpy as np
from math import sqrt, sin, cos, tan, asin, acos, atan, pi

def _plane_distance(x1, y1, x2, y2):
    return sqrt((x2 - x1)**2 + (y2 - y1)**2)

def plane_distance(xs1, ys1, xs2, ys2):
    if hasattr(xs1, "__len__"):
        return np.array([_plane_distance(x1, y1, x2, y2)
                        for x1, y1, x2, y2 in zip(xs1, ys1, xs2, ys2)])
    else:
        return _plane_distance(xs1, ys1, xs2, ys2)

def _sphere_distance(lon1, lat1, lon2, lat2, radius):
    dx = abs(lon1-lon2)
    dy = abs(lat1-lat2)

    if dx > 0.01 or dy > 0.01:
        # use spherical law of cosines
        d_ = acos(sin(lat1) * sin(lat2) + cos(lat1) * cos(lat2) * cos(dx))

    else:
        # use haversine
        d_ = (2 * asin(sqrt(sin(dy / 2.)**2 +
                        cos(lat1) * cos(lat2) * sin(dx / 2.)**2)))
    return radius * d_

def sphere_distance(lons1, lats1, lons2, lats2, radius):
    if hasattr(lons1, "__len__"):
        return np.array([_sphere_distance(ln1, lt1, ln2, lt2, radius)
                        for ln1, lt1, ln2, lt2
                        in zip(lons1, lats1, lons2, lats2)])
    else:
        return _sphere_distance(lons1, lats1, lons2, lats2, radius)


def _plane_azimuth(x1, y1, x2, y2):
    """ Return cartesian azimuth between points (scalar func) """
    dx = x2 - x1
    dy = y2 - y1
    if dx > 0:
        if dy > 0:
            return np.arctan(dx/dy)
        elif dy < 0:
            return np.pi - np.arctan(-dx/dy)
        else:
            return 0.5*np.pi
    elif dx < 0:
        if dy > 0:
            return 2*np.pi - np.arctan(-dx/dy)
        elif dy < 0:
            return np.pi + np.arctan(dx/dy)
        else:
            return 1.5*np.pi
    else:
        if dy > 0:
            return 0.0
        else:
            return np.pi

def plane_azimuth(xs1, ys1, xs2, ys2):
    """ Return cartesian azimuth between points """
    if hasattr(xs1, "__len__"):
        return np.array([_plane_azimuth(x1, y1, x2, y2)
                        for x1, y1, x2, y2 in zip(xs1, ys1, xs2, ys2)])
    else:
        return _plane_azimuth(xs1, ys1, xs2, ys2)

def _sphere_azimuth(lon1, lat1, lon2, lat2):
    dlon = lon2 - lon1
    if (cos(lat1) * tan(lat2) - sin(lat1) * cos(dlon)) == 0:
        az = 0.5*pi
    else:
        az = atan(sin(dlon) / (cos(lat1) * tan(lat2) - sin(lat1) * cos(dlon)))
    
    if unroll_angle(dlon) <= pi:
        if lat2 < lat1:
            az = az + pi
    else:
        if lat1 < lat2:
            az = az + 2*pi
        else:
            az = az + pi
    return unroll_angle(az)

def sphere_azimuth(lons1, lats1, lons2, lats2):
    if hasattr(lons1, "__len__"):
        return np.array([_sphere_azimuth(lon1, lat1, lon2, lat2)
                        for lon1, lat1, lon2, lat2
                        in zip(lons1, lats1, lons2, lats2)])
    else:
        return _sphere_azimuth(lons1, lats1, lons2, lats2)


###### Utility functions ######
def unroll_angle(alpha):
    if hasattr(alpha, "__len__"):
        alpha = np.array([unroll_angle(a) for a in alpha])
    else:
        while alpha < 0:
            alpha += 2*pi
        while alpha >= 2*pi:
            alpha -= 2*pi
    return alpha




from .vector import Point, Multipoint, Line, Polygon
from .crs import LonLatWGS84

us_capitols = [Point((-112.1, 33.57), properties={"n": "Phoenix, Arizona, United States"}, crs=LonLatWGS84),
               Point((-121.5, 38.57), properties={"n": "Sacramento, California, United States"}, crs=LonLatWGS84),
               Point((-84.42, 33.76), properties={"n": "Atlanta, Georgia, United States"}, crs=LonLatWGS84),
               Point((-86.15, 39.78), properties={"n": "Indianapolis, Indiana, United States"}, crs=LonLatWGS84),
               Point((-112.0, 46.60) , properties={"n": "Helena, Montana, United States"}, crs=LonLatWGS84),
               Point((-82.99, 39.98), properties={"n": "Columbus, Ohio, United States"}, crs=LonLatWGS84),
               Point((-77.48, 37.53), properties={"n": "Richmond, Virginia, United States"}, crs=LonLatWGS84),
               Point((-95.69, 39.04), properties={"n": "Topeka, Kansas, United States"}, crs=LonLatWGS84),
               Point((-71.02, 42.33), properties={"n": "Boston, Massachusetts, United States"}, crs=LonLatWGS84),
               Point((-96.68, 40.81), properties={"n": "Lincoln, Nebraska, United States"}, crs=LonLatWGS84),
               Point((-97.51, 35.47), properties={"n": "Oklahoma City, Oklahoma, United States"}, crs=LonLatWGS84),
               Point((-134.2, 58.37), properties={"n": "Juneau, Alaska, United States"}, crs=LonLatWGS84),
               Point((-100.3, 44.38), properties={"n": "Pierre, South Dakota, United States"}, crs=LonLatWGS84),
               Point((-157.8, 21.31), properties={"n": "Honolulu, Hawaii, United States"}, crs=LonLatWGS84),
               Point((-86.27, 32.35), properties={"n": "Montgomery, Alabama, United States"}, crs=LonLatWGS84),
               Point((-92.36, 34.73), properties={"n": "Little Rock, Arkansas, United States"}, crs=LonLatWGS84),
               Point((-104.9, 39.76), properties={"n": "Denver, Colorado, United States"}, crs=LonLatWGS84),
               Point((-72.68, 41.77), properties={"n": "Hartford, Connecticut, United States"}, crs=LonLatWGS84),
               Point((-75.52, 39.16), properties={"n": "Dover, Delaware, United States"}, crs=LonLatWGS84),
               Point((-77.02, 38.90) , properties={"n": "Washington, District of Columbia, United States"}, crs=LonLatWGS84),
               Point((-84.25, 30.46), properties={"n": "Tallahassee, Florida, United States"}, crs=LonLatWGS84),
               Point((-116.2, 43.60) , properties={"n": "Boise, Idaho, United States"}, crs=LonLatWGS84),
               Point((-89.67, 39.76), properties={"n": "Springfield, Illinois, United States"}, crs=LonLatWGS84),
               Point((-93.62, 41.57), properties={"n": "Des Moines, Iowa, United States"}, crs=LonLatWGS84),
               Point((-84.87, 38.19), properties={"n": "Frankfort, Kentucky, United States"}, crs=LonLatWGS84),
               Point((-91.13, 30.45), properties={"n": "Baton Rouge, Louisiana, United States"}, crs=LonLatWGS84),
               Point((-69.73, 44.33), properties={"n": "Augusta, Maine, United States"}, crs=LonLatWGS84),
               Point((-76.51, 38.97), properties={"n": "Annapolis, Maryland, United States"}, crs=LonLatWGS84),
               Point((-84.56, 42.71), properties={"n": "Lansing, Michigan, United States"}, crs=LonLatWGS84),
               Point((-93.10, 44.95), properties={"n": "Saint Paul, Minnesota, United States"}, crs=LonLatWGS84),
               Point((-90.21, 32.32), properties={"n": "Jackson, Mississippi, United States"}, crs=LonLatWGS84),
               Point((-92.18, 38.57), properties={"n": "Jefferson City, Missouri, United States"}, crs=LonLatWGS84),
               Point((-119.7, 39.15), properties={"n": "Carson City, Nevada, United States"}, crs=LonLatWGS84),
               Point((-71.56, 43.23), properties={"n": "Concord, New Hampshire, United States"}, crs=LonLatWGS84),
               Point((-74.76, 40.22), properties={"n": "Trenton, New Jersey, United States"}, crs=LonLatWGS84),
               Point((-106.0, 35.67), properties={"n": "Santa Fe, New Mexico, United States"}, crs=LonLatWGS84),
               Point((-73.80, 42.67), properties={"n": "Albany, New York, United States"}, crs=LonLatWGS84),
               Point((-78.64, 35.83), properties={"n": "Raleigh, North Carolina, United States"}, crs=LonLatWGS84),
               Point((-100.8, 46.81), properties={"n": "Bismarck, North Dakota, United States"}, crs=LonLatWGS84),
               Point((-123.0, 44.92), properties={"n": "Salem, Oregon, United States"}, crs=LonLatWGS84),
               Point((-76.89, 40.28), properties={"n": "Harrisburg, Pennsylvania, United States"}, crs=LonLatWGS84),
               Point((-71.42, 41.82), properties={"n": "Providence, Rhode Island, United States"}, crs=LonLatWGS84),
               Point((-81.01, 34.02), properties={"n": "Columbia, South Carolina, United States"}, crs=LonLatWGS84),
               Point((-86.78, 36.17), properties={"n": "Nashville, Tennessee, United States"}, crs=LonLatWGS84),
               Point((-97.76, 30.31), properties={"n": "Austin, Texas, United States"}, crs=LonLatWGS84),
               Point((-111.9, 40.78), properties={"n": "Salt Lake City, Utah, United States"}, crs=LonLatWGS84),
               Point((-72.57, 44.27), properties={"n": "Montpelier, Vermont, United States"}, crs=LonLatWGS84),
               Point((-122.9, 47.04), properties={"n": "Olympia, Washington, United States"}, crs=LonLatWGS84),
               Point((-81.63, 38.35), properties={"n": "Charleston, West Virginia, United States"}, crs=LonLatWGS84),
               Point((-89.43, 43.09), properties={"n": "Madison, Wisconsin, United States"}, crs=LonLatWGS84),
               Point((-104.8, 41.15), properties={"n": "Cheyenne, Wyoming, United States"}, crs=LonLatWGS84)]



from ez_setup import use_setuptools
use_setuptools()
from os.path import exists
from setuptools import setup, Extension
import numpy

VERSION = "0.4.3"

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
    include_dirs = [numpy.get_include()]
    ext = ".pyx"
except ImportError:
    USE_CYTHON = False
    include_dirs = []
    ext = ".c"
    print("Warning: Cython not imported")
    print("If C sources exist in the source tree, they will be used")

extensions = [Extension("karta.raster.crfuncs",
                        ["karta/raster/crfuncs"+ext],
                        include_dirs=include_dirs),
              Extension("karta.vector._cvectorgeo",
                        ["karta/vector/_cvectorgeo"+ext],
                        include_dirs=include_dirs)]

if USE_CYTHON:
    extensions = cythonize(extensions)

for extension in extensions:
    if not all(exists(src) for src in extension.sources):
        # C-extension sources missing, so don't try to build them
        extensions = []
        print("Warning: Extension source not found: {0}".format(extension.sources))
        print("Not building accelerated modules")
        break

setup(
    name = "karta",
    version = VERSION,
    install_requires = ["numpy>=1.6", "cython>=0.15", "pyproj>=1.9", "pyshp>=1.2"],
    author = "Nat Wilson",
    author_email = "njwilson23@gmail.com",
    packages = ["karta", "karta.vector", "karta.raster"],
    url = "http://www.ironicmtn.com/karta.html",
    description = "Geospatial analysis in Python",
    long_description = """
Karta - tidy package for geospatial computation
===============================================

*Karta* is a simple and fast framework for spatial analysis in Python.

Components:

- Clean geographically-aware vector and gridded data types
- Integration with pyproj to support a wide range of coordinate systems and
  transformations
- A selection of geographical analysis methods including geodetic length and
  area calculations, intersections, convex hulls, raster sampling and profiling,
  and grid warping
- IO for several common geographical formats, including GeoJSON, shapefiles
  (through pyshp), ESRI ASCII, and GeoTiff (through GDAL). Vector geometries
  implement ``__geo_interface__``.

*Karta* works with Python 2.6+ and Python 3.3+.

DOCUMENTATION
-------------

See the `online manual <http://www.ironicmtn.com/kartadocs/karta-manual.html>`_,
read the tutorial_, or search the `API documentation`_.

.. _tutorial: http://www.ironicmtn.com/kartadocs/tutorial.html
.. _API documentation: http://www.ironicmtn.com/kartadocs/reference.html

The manual can also be built offline with Sphinx by running ``make`` from the
``doc/`` directory. The documentation is built from source code docstrings and
information in the `Wiki <https://github.com/njwilson23/karta/wiki/Tutorial>`_.
""",
    download_url = "https://github.com/njwilson23/karta/archive/master.zip",
    classifiers = ["Programming Language :: Python",
                   "Programming Language :: Python :: 3",
                   "Development Status :: 4 - Beta",
                   "Topic :: Scientific/Engineering"],
    license = "MIT License",
    ext_modules = extensions
)




"""
GPX IO and manipulation (UNDER DEVELOPMENT)

To do:
    - XML namespaces aren't really handled properly
    - metadata node not addressed
"""

import collections
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import ElementTree, Element
from xml.dom import minidom

Point = collections.namedtuple("Point", ["lonlat", "properties", "extensions"])
Trkseg = collections.namedtuple("Trkseg", ["trkpts", "properties", "extensions"])
Track = collections.namedtuple("Track", ["trksegs", "properties", "extensions"])
Route = collections.namedtuple("Route", ["rtepts", "properties", "extensions"])

ns = "{http://www.topografix.com/GPX/1/1}"


VALID_PROPERTIES = ("ele", "time", "magvar", "geoidheight", "name", "cmt",
                    "desc", "src", "link", "sym", "type", "fix", "sat", "hdop",
                    "vdop", "pdop", "ageofdgpsdata", "dgpsid")

def strip_namespace(s):
    return s[s.index("}")+1:]

class GPX(object):
    """ Represents a GPX document, with an internal representation of
    waypoints, tracks, and route that loosely approximates the XML structure.
    Provides methods to easily add Point-like and Line-like objects as GPX
    types. """

    def __init__(self, f=None, waypoints=None, tracks=None, routes=None):
        """ Create a GPX object, either from a GPX file or from lists of
        waypoints, tracks, and routes. """

        self.waypts = []
        self.tracks = []
        self.routes = []

        if f is not None:
            self.fromfile(f)

        else:
            if waypoints is not None:
                for waypt in waypoints:
                    self.add_waypoint(waypt)
            if tracks is not None:
                for track in tracks:
                    self.add_track(track)
            if routes is not None:
                for route in routes:
                    self.add_route(route)
        return

    def _readextensions(self, node):
        extensions = {}
        try:
            for ext in node.find(ns + "extensions"):
                extensions[strip_namespace(ext.tag)] = ext.text
        except TypeError:
            pass
        return extensions

    def _readproperties(self, node, exclude=()):
        properties = {}
        for subnode in node:
            tag = strip_namespace(subnode.tag)
            if tag not in exclude:
                properties[tag] = subnode.text
        return properties

    def _readwpt(self, wpt):
        properties = self._readproperties(wpt, exclude=("extensions",))
        extensions = self._readextensions(wpt)
        lon = wpt.attrib["lon"]
        lat = wpt.attrib["lat"]
        return Point((lon, lat), properties, extensions)

    def _dict2gpx(self, parent, properties):
        for p in properties:
            sub = Element(ns + p)
            sub.text = str(properties[p])
            parent.append(sub)
        return parent

    def _extensions2gpx(self, parent, extensions):
        ext = Element(ns + "extensions")
        ext = self._dict2gpx(ext, extensions)
        parent.append(ext)
        return parent

    def _build_gpx_wpt(self, waypt, tag="wpt"):
        """ Build <wpt> node. """
        wpt = Element(ns + tag, lon=str(waypt.lonlat[0]), lat=str(waypt.lonlat[1]))
        wpt = self._dict2gpx(wpt, waypt.properties)
        wpt = self._extensions2gpx(wpt, waypt.extensions)
        return wpt

    def _build_gpx_trk(self, track):
        """ Build "trk" nodes. """
        trk = Element(ns + "trk")
        trk = self._dict2gpx(trk, track.properties)
        trk = self._extensions2gpx(trk, track.extensions)

        for segment in track.trksegs:
            trkseg = Element(ns + "trkseg")
            trkseg = self._dict2gpx(trkseg, segment.properties)
            trkseg = self._extensions2gpx(trkseg, segment.extensions)

            for trackpt in segment.trkpts:
                trkpt = self._build_gpx_wpt(trackpt, tag="trkpt")
                trkseg.append(trkpt)

            trk.append(trkseg)
        return trk

    def _build_gpx_rte(self, route):
        rte = Element(ns + "rte")
        rte = self._dict2gpx(rte, route.properties)
        rte = self._extensions2gpx(rte, route.extensions)
        for routept in route.rtepts:
            rte.append(self._build_gpx_wpt(routept, tag="rtept"))
        return rte

    def fromfile(self, f):
        """ Read a GPX document from *f*, which may be a filename or a
        file-like object. """

        gpxtree = ElementTree(file=f)
        self.gpx = gpxtree.getroot()

        for node in self.gpx.findall(ns + "wpt"):
            self.parse_wpt(node)

        for node in self.gpx.findall(ns + "trk"):
            self.parse_trk(node)

        for node in self.gpx.findall(ns + "rte"):
            self.parse_rte(node)

        return

    def parse_wpt(self, wpt):
        """ Parse a <wpt> node, updating self.waypoints. """
        point = self._readwpt(wpt)
        name = wpt.properties.get("name", "waypoint_" + str(len(self.waypoints)))
        self.waypoints[name] = point
        return

    def parse_trk(self, trk):
        """ Parse a <trk> node, updating self.tracks. """
        segments = []
        trkproperties = self._readproperties(trk, exclude=("trkseg",))
        trkextensions = self._readextensions(trk)

        for trkseg in trk.findall(ns + "trkseg"):

            points = [self._readwpt(trkpt) for trkpt in trkseg.findall(ns + "trkpt")]
            properties = self._readproperties(trkseg, exclude=("trkpt",))
            extensions = self._readextensions(trkseg)
            segments.append(Trkseg(points, properties, extensions))

        self.tracks.append(Track(segments, trkproperties, trkextensions))
        return

    def parse_rte(self, rte):
        properties = self._readproperties(rte, exclude=("rtept",))
        extensions = self._readextensions(rte)
        points = [self._readwpt(rtept) for rtept in rte.findall(ns + "rtept")]
        self.routes.append(Route(points, properties, extensions))
        return

    def add_waypoint(self, waypoint):
        """ Add a Point-like object as a waypoint. Properties and extension
        types are taken from waypoint.properties attribute. """
        properties = {}
        extensions = {}
        for key in waypoint.properties:
            if key in VALID_PROPERTIES:
                properties[key] = str(waypoint.properties[key])
            else:
                extensions[key] = str(waypoint.properties[key])
        waypt = Point(waypoint.vertex, properties, extensions)
        self.waypts.append(waypt)
        return

    def add_track(self, track, attributes=None):
        """ Add a list of Line-like objects as a track. Dictionaries of
        properties and extension types for the track are accepted as keyword
        arguments.

        Properties and extension types for the track segments are taken from
        the `properties` attribute of each Line-like object.

        Properties and extensions types for each track point are taken from the
        `data` attribute of each Line-like object.

        INCOMPLETE:

        Needs to distinguish between properties and extensions at the trkseg
        and trkpt levels

        Needs to properly add and <ele> property for Lines of rank 3
        """
        segments = []
        properties = {}
        extensions = {}

        if attributes is not None:
            for key in attributes:
                if key in VALID_PROPERTIES:
                    properties[key] = str(attributes[key])
                else:
                    properties[key] = str(attributes[key])

        for line in track:
            points = []
            sg_properties = {}
            sg_extensions = {}

            for key in line.properties:
                if key in VALID_PROPERTIES:
                    sg_properties[key] = str(line.properties[key])
                else:
                    sg_extensions[key] = str(line.properties[key])

            if line.data is not None:
                pt_properties = [k for k in line.data.keys() if k in VALID_PROPERTIES]
                pt_extensions = [k for k in line.data.keys() if k not in VALID_PROPERTIES]
            else:
                pt_properties = []
                pt_extensions = []

            for i, xy in enumerate(line.vertices):

                ptprop = {}
                ptexte = {}
                for key in pt_properties:
                    ptprop[key] = line.data[key][i]
                    if len(xy) == 3:
                        ptprop["ele"] = xy[2]
                for key in pt_extensions:
                    ptexte[key] = line.data[key][i]

                points.append(Point((xy[0], xy[1]), ptprop, ptexte))

            segments.append(Trkseg(points, sg_properties, sg_extensions))

        self.tracks.append(Track(segments, properties, extensions))
        return

    def add_route(self, route):
        """ Add a list of Line-like objects as a route. Properties and
        extension types for the route are taken from the `properties` attribute
        of the Line-like object.

        Properties and extensions types for each route point are taken from the
        `data` attribute of each Line-like object.

        INCOMPLETE:

        Needs to distinguish between properties and extensions at the trkseg
        and trkpt levels

        Needs to properly add and <ele> property for Lines of rank 3
        """
        points = []
        for i, vertex in enumerate(route.vertices):
            prop = {}
            if route.data is not None:
                for k in rout.data.fields:
                    prop[k] = route.data[k][i]
            points.append(Point((vertex[0], vertex[1]), prop, {}))
        self.routes.append(Route(points, route.properties, {}))
        return

    def writefile(self, fnm, waypts=True, tracks=True, routes=True):
        """ Write GPX object to a GPX file. Writes all waypoints, tracks, and
        routes by default, which can be changed by changing the kwargs to
        False. """
        gpx = Element(ns + "gpx", version="1.1", creator="karta")

        if waypts:
            for waypt in self.waypts:
                gpx.append(self._build_gpx_wpt(waypt))
        if tracks:
            for track in self.tracks:
                gpx.append(self._build_gpx_trk(track))
        if routes:
            for route in self.routes:
                gpx.append(self._build_gpx_rte(route))

        xmlstring = ET.tostring(gpx)
        output = minidom.parseString(xmlstring).toprettyxml(indent="  ")
        with open(fnm, "w") as f:
            f.write(output)
        return




# -*- coding: utf-8 -*-
""" Coordinate reference systems

Implements CRS classes for different kinds of spatial reference systems:

    - Cartesian
    - GeographicalCRS
    - Spherical
    - Proj4CRS
"""

import numpy as np
import pyproj
from . import geodesy

# A CRS class needs to have pyproj Proj and Geod instances. Exceptions are
# `Cartesian` and `Spherical` which are implemented specially.

class CRS(object):
    """ Base class for coordinate system instances
    
    Subclasses should at a minimum define a *name* attribute. Providing
    *project* *forward* and *inverse* methods permits a geometry associated with
    a CRS subclass to be used in analyses.

    *project* transforms between world and projected coordinates, and takes the
    optional boolean parameter *inverse* (default `False`)

    *forward* takes world coordinates, an azimuth, and a distance, and returns
    world coordinates

    *inverse* takes two pairs of world coordinates and returns an azimuth, a
    back azimuth, and a distance
    """
    def __str__(self):
        return "<CRS {0}>".format(self.name)

class Cartesian(CRS):
    """ Cartiesian (flat-earth) reference systems with (x, y) coordinates """
    name = "Cartesian"

    @staticmethod
    def project(x, y, inverse=False):
        return x, y

    @staticmethod
    def forward(x, y, az, dist, radians=False):
        """ Returns x, y, and back azimuths """
        if not radians:
            az = np.array(az) / 180 * np.pi

        x2 = x + dist * np.sin(az)
        y2 = y + dist * np.cos(az)
        baz = geodesy.unroll_angle(az + np.pi)

        if not radians:
            baz = np.array(baz) * 180 / np.pi
        return x2, y2, baz

    @staticmethod
    def inverse(x1, y1, x2, y2, radians=False):
        """ Returns forward and back azimuths and distances """
        dist = geodesy.plane_distance(x1, y1, x2, y2)
        az = geodesy.plane_azimuth(x1, y1, x2, y2)
        baz = geodesy.unroll_angle(az + np.pi)

        if not radians:
            az = az * 180 / np.pi
            baz = baz * 180 / np.pi
        return az, baz, dist

class GeographicalCRS(CRS):
    """ Reference systems with longitude-latitude (θ, φ) coordinates """
    def __init__(self, spheroid, name):
        self._geod = pyproj.Geod(spheroid)
        self.name = name
        return

    @staticmethod
    def project(x, y, inverse=False, radians=False):
        if not radians:
            return x, y
        else:
            return np.array(x)/180 * np.pi, np.array(y)/180 * np.pi

    def forward(self, *args, **kwargs):
        return self._geod.fwd(*args, **kwargs)

    def inverse(self, *args, **kwargs):
        return self._geod.inv(*args, **kwargs)


class Spherical(GeographicalCRS):
    """ Spherical (θ, φ) coordinate system. """
    name = "Spherical"

    def __init__(self, radius):
        self.radius = radius

    def forward(self, lons, lats, az, dist, radians=False):
        """ Returns lons, lats, and back azimuths """
        if not radians:
            lons = np.array(lons) * np.pi / 180.0
            lats = np.array(lats) * np.pi / 180.0
            az = np.array(az) * np.pi / 180.0


        d_ = dist / self.radius
        lats2 = np.arcsin(np.sin(lats) * np.cos(d_) + 
                    np.cos(lats) * np.sin(d_) * np.cos(az))
        dlons = np.arccos((np.cos(d_) - np.sin(lats2) * np.sin(lats)) /
                          (np.cos(lats) * np.cos(lats2)))
        baz = np.arccos((np.sin(lats) - np.cos(d_) * np.sin(lats2)) /
                        (np.sin(d_) * np.cos(lats2)))
        if 0 <= az < np.pi:
            lons2 = lons + dlons
            baz = -baz
        elif np.pi <= az < 2*np.pi:
            lons2 = lons - dlons
        else:
            raise ValueError("azimuth should be [0, 2pi)")

        baz = geodesy.unroll_angle(baz)

        if not radians:
            lons2 = np.array(lons2) * 180 / np.pi
            lats2 = np.array(lats2) * 180 / np.pi
            baz = np.array(baz) * 180 / np.pi
        return lons2, lats2, baz

    def inverse(self, lons1, lats1, lons2, lats2, radians=False):
        """ Returns forward and back azimuths and distances """
        if not radians:
            lons1 = lons1 * np.pi / 180.0
            lons2 = lons2 * np.pi / 180.0
            lats1 = lats1 * np.pi / 180.0
            lats2 = lats2 * np.pi / 180.0

        az = geodesy.sphere_azimuth(lons1, lats1, lons2, lats2)
        baz = geodesy.sphere_azimuth(lons2, lats2, lons1, lats1)
        dist = geodesy.sphere_distance(lons1, lats1, lons2, lats2, self.radius)

        if not radians:
            az = az * 180 / np.pi
            baz = baz * 180 / np.pi
        return az, baz, dist

class Proj4CRS(CRS):
    """ Custom reference systems, which may be backed by a *pypoj.Proj* instance
    or a custom projection function """

    def __init__(self, proj, spheroid, name=None):
        self.project = pyproj.Proj(proj)
        self._geod = pyproj.Geod(spheroid)

        if name is not None:
            self.name = name
        else:
            self.name = proj
        return

    def __eq__(self, other):
        if (getattr(self.project, "srs", 0) == getattr(other.project, "srs", 1) and
            getattr(self._geod, "initstring", 0) == getattr(other._geod, "initstring", 1)):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def forward(self, *args, **kwargs):
        return self._geod.fwd(*args, **kwargs)

    def inverse(self, *args, **kwargs):
        return self._geod.inv(*args, **kwargs)

class CRSError(Exception):
    """ Exception to raise for invalid geodetic operations. """
    def __init__(self, message=''):
        self.message = message

############ Predefined CRS instances ############


SphericalEarth = Spherical(6371.0)
LonLatWGS84 = GeographicalCRS("+ellps=WGS84", "WGS84 (Geographical)")
LonLatNAD27 = GeographicalCRS("+ellps=clrk66", "NAD27 (Geographical)")
LonLatNAD83 = GeographicalCRS("+ellps=GRS80", "NAD83 (Geographical)")

UPSNorth = Proj4CRS(proj="+proj=stere +lat_0=90 +lat_ts=90 +lon_0=0 +k=0.994 +x_0=2000000 +y_0=2000000 +units=m +datum=WGS84 +no_defs",
        spheroid="+ellps=WGS84", name="Universal Polar Stereographic (North)")

UPSSouth = Proj4CRS(proj="+proj=stere +lat_0=-90 +lat_ts=-90 +lon_0=0 +k=0.994 +x_0=2000000 +y_0=2000000 +units=m +datum=WGS84 +no_defs",
        spheroid="+ellps=WGS84", name="Universal Polar Stereographic (South)")

NSIDCNorth = Proj4CRS(proj="+proj=stere +lat_0=90 +lat_ts=70 +lon_0=-45 +k=1 +x_0=0 +y_0=0 +units=m +datum=WGS84 +no_defs",
        spheroid="+ellps=WGS84", name="NSIDC (North)")

NSIDCSouth = Proj4CRS(proj="+proj=stere +lat_0=-90 +lat_ts=-70 +lon_0=0 +k=1 +x_0=0 +y_0=0 +units=m +datum=WGS84 +no_defs",
        spheroid="+ellps=WGS84", name="NSIDC (South)")





""" Functions for reading raster data sources as RegularGrid objects """
import numpy as np
from .grid import RegularGrid
from ..crs import Proj4CRS, GeographicalCRS

try:
    from . import _gtiff
    HAS_GDAL = True
except ImportError:
    HAS_GDAL = False

from . import _aai
# from . import _dem

def read_aai(fnm):
    """ Convenience function to open a ESRI ASCII grid and return a RegularGrid
    instance.
    """
    values, aschdr = _aai.aairead(fnm)
    t = {'xllcenter': aschdr['xllcorner'] + 0.5 * aschdr['cellsize'],
         'yllcenter': aschdr['yllcorner'] + 0.5 * aschdr['cellsize'],
         'dx'       : aschdr['cellsize'],
         'dy'       : aschdr['cellsize'],
         'xrot'     : 0.0,
         'yrot'     : 0.0}
    values[values==aschdr['nodata_value']] = np.nan
    return RegularGrid(t, values=values[::-1])

def proj4_isgeodetic(s):
    return ("lonlat" in s) or ("longlat" in s) or \
            ("latlon" in s) or ("latlong" in s)

def read_gtiff(fnm, band=1):
    """ Convenience function to open a GeoTIFF and return a RegularGrid
    instance.

    Parameters
    ----------

    fnm : GeoTiff file path

    band : band to open (default 1)
    """
    if not HAS_GDAL:
        raise ImportError("Reading GeoTiffs depends on GDAL, which could not "
                          "be imported")
    arr, hdr = _gtiff.read(fnm, band)
    t = {'xllcenter'  : hdr['xulcorner'] + 0.5 * (hdr['dx'] + hdr['sx']),
         'yllcenter'  : hdr['yulcorner'] + (hdr['ny'] - 0.5) * hdr['dy'] - 0.5 * hdr['sy'],
         'dx'         : hdr['dx'],
         'dy'         : -hdr['dy'],
         'xrot'       : hdr['sx'],
         'yrot'       : -hdr['sy']}

    geodstr = "+a={a} +f={f}".format(a=hdr["srs"]["semimajor"],
                                     f=hdr["srs"]["flattening"])
    if proj4_isgeodetic(hdr["srs"]["proj4"]):
        crs = GeographicalCRS(geodstr, name="Imported GTiff")
    else:
        crs = Proj4CRS(hdr["srs"]["proj4"], geodstr, name="Imported GTiff")
    return RegularGrid(t, values=arr.squeeze()[::-1], crs=crs)

# Aliases for backwards compat.
gtiffread = read_gtiff
aairead = read_aai

